"""
    This class interpolates dataframe positions based on Datetime.
    It provides the user with the flexibility to use linear or cubic interpolation.
    In general, the user passes the dataframe, time jum and the interpolation type,
    based on the type the proper function is mapped. And if the time difference
    exceeds the time jump, the interpolated point is added to the position with large jump
    with a time increase of time jump. This interpolated row is added to the dataframe.

    | Authors: Yaksh J Haranwala, Salman Haidri

"""
import itertools
import multiprocessing as mlp
import os
from math import ceil
from typing import Optional, Text, Union

import pandas
import pandas as pd

from ptrail.core.TrajectoryDF import PTRAILDataFrame as NumTrajDF
from ptrail.preprocessing.helpers import Helpers as helper
from ptrail.utilities import constants as const

num = os.cpu_count()
NUM_CPU = ceil((num * 2) / 3)


class Interpolation:
    @staticmethod
    def interpolate_position(dataframe: NumTrajDF, time_jump: float, ip_type: Optional[Text] = 'linear'):
        """
            Interpolate the position of an object and create new points using one of
            the interpolation methods provided by the Library. Currently, the library
            supports the following 4 interpolation methods:

                1. Linear Interpolation
                2. Cubic-Spline Interpolation
                3. Kinematic Interpolation
                4. Random Walk Interpolation

            Warning
            -------
                The Interpolation methods will only return the 4 mandatory library columns
                because it is not possible to interpolate other data that may or may not be
                present in the dataset apart from latitude, longitude and datetime. As a
                result, other columns are dropped.

            Note
            ----
                The time-jump parameter specifies where the new points are to be
                inserted based on the time difference between 2 consecutive points.
                However, it does not guarantee that the dataset will be brought down
                to having difference between 2 consecutive points equal to or
                less than the user specified time jump.

            Note
            ----
                The time-jump is specified in seconds. Hence, if the user-specified
                time-jump is not sensible, then the execution of the method will take
                a very long time.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe containing the original dataset.
                time_jump: float
                    The maximum time difference between 2 consecutive points.
                ip_type: Optional[Text], default = linear
                    The type of interpolation that is to be used.

            Returns
            -------
                PTRAILDataFrame:
                    The dataframe containing the interpolated trajectory points.
        """
        # First, lets split the dataframe into smaller chunks containing
        # points of only n trajectory per chunk. n is
        df = dataframe.reset_index()
        df_chunks = helper._df_split_helper(df)

        # Create a pool of processes which has number of processes
        # equal to the number of unique dataframe partitions.
        processes = [None] * len(df_chunks)
        manager = mlp.Manager()
        return_list = manager.list()

        ip_type = ip_type.lower().strip()
        if ip_type == 'linear':
            for i in range(len(processes)):
                processes[i] = mlp.Process(target=Interpolation._linear_ip,
                                           args=(df_chunks[i], time_jump, return_list))
                processes[i].start()

            for j in range(len(processes)):
                processes[j].join()

        elif ip_type == 'cubic':
            for i in range(len(processes)):
                processes[i] = mlp.Process(target=Interpolation._cubic_ip,
                                           args=(df_chunks[i], time_jump, return_list))
                processes[i].start()

            for j in range(len(processes)):
                processes[j].join()
        elif ip_type == 'kinematic':
            for i in range(len(processes)):
                processes[i] = mlp.Process(target=Interpolation._kinematic_ip,
                                           args=(df_chunks[i], time_jump, return_list))
                processes[i].start()

            for j in range(len(processes)):
                processes[j].join()
        elif ip_type == 'random-walk':
            for i in range(len(processes)):
                processes[i] = mlp.Process(target=Interpolation._random_walk_ip,
                                           args=(df_chunks[i], time_jump, return_list))
                processes[i].start()

            for j in range(len(processes)):
                processes[j].join()
        else:
            raise ValueError(f"Interpolation type: {ip_type} specified does not exist. Please check the"
                             "interpolation type specified and type again.")

        return NumTrajDF(pd.concat(return_list).reset_index(),
                         const.LAT, const.LONG, const.DateTime, const.TRAJECTORY_ID)

    @staticmethod
    def _linear_ip(dataframe: Union[pd.DataFrame, NumTrajDF], time_jump: float, return_list: list):
        """
            Interpolate the position of points using the Linear Interpolation method. It makes
            the use of numpy's interpolation technique for the interpolation of the points.

            WARNING: Do not use this method directly as it will run slower. Instead,
                     use the method interpolate_position() and specify the ip_type as
                     linear to perform linear interpolation much faster.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe containing the original data.
                time_jump: float
                    The maximum time difference between 2 points. If the time difference between
                    2 consecutive points is greater than the time jump, then another point will
                    be inserted between the given 2 points.
                return_list: list
                    The list used by the Multiprocessing manager to get the return values

            Returns
            -------
                pandas.core.dataframe.DataFrame:
                    The dataframe enhanced with interpolated points.
        """
        # First, reset the index, extract the Latitude, Longitude, DateTime and Trajectory ID columns
        # and set the DateTime column only as the index. Then, store all the unique Trajectory IDs in
        # a list.
        dataframe = dataframe.reset_index()[
            [const.DateTime, const.TRAJECTORY_ID, const.LAT, const.LONG]].set_index(const.DateTime)

        # Split the smaller dataframe further into smaller chunks containing only 1
        # Trajectory ID per index.
        ids_ = list(dataframe[const.TRAJECTORY_ID].value_counts().keys())
        df_chunks = [dataframe.loc[dataframe[const.TRAJECTORY_ID] == ids_[i]] for i in range(len(ids_))]

        # Here, create as many processes at once as there are number of CPUs available in
        # the system - 1. One CPU is kept free at all times in order to not block up
        # the system. (Note: The blocking of system is mostly prevalent in Windows and does
        # not happen very often in Linux. However, out of caution 1 CPU is kept free regardless
        # of the system.)
        small_pool = mlp.Pool(NUM_CPU)
        final = small_pool.starmap(helper.linear_help,
                                   zip(df_chunks, ids_, itertools.repeat(time_jump)))
        small_pool.close()
        small_pool.join()

        # Append the smaller dataframe to process manager list so that result
        # can be finally merged into a larger dataframe.
        return_list.append(pd.concat(final))

    @staticmethod
    def _cubic_ip(dataframe: Union[pd.DataFrame, NumTrajDF], time_jump: float, return_list: list):
        try:
            """
                Method for cubic interpolation of a dataframe based on the time jump provided.
                It makes use of scipy library's CubicSpline functionality and interpolates
                the coordinates based on the Datetime of the dataframe.
    
                WARNING: Do not use this method directly as it will run slower. Instead,
                         use the method interpolate_position() and specify the ip_type as
                         cubic to perform cubic interpolation much faster.
    
                Parameters
                ----------
                    dataframe: Union[pd.DataFrame, NumTrajDF]
                        The dataframe on which interpolation is to be performed
                    time_jump: float
                        The maximum time difference allowed to have between rows
                    return_list: list
                        The list used by the Multiprocessing manager to get the return values
    
                Returns
                -------
                    pandas.core.dataframe.DataFrame:
                        The dataframe containing the new interpolated points.
    
            """
            # First, reset the index, extract the Latitude, Longitude, DateTime and Trajectory ID columns
            # and set the DateTime column only as the index. Then, store all the unique Trajectory IDs in
            # a list.
            dataframe = dataframe.reset_index()[
                [const.DateTime, const.TRAJECTORY_ID, const.LAT, const.LONG]].set_index(const.DateTime)

            # Split the smaller dataframe further into smaller chunks containing only 1
            # Trajectory ID per index.
            ids_ = list(dataframe[const.TRAJECTORY_ID].value_counts().keys())
            df_chunks = [dataframe.loc[dataframe[const.TRAJECTORY_ID] == ids_[i]] for i in range(len(ids_))]

            # Here, create as many processes at once as there are number of CPUs available in
            # the system - 1. One CPU is kept free at all times in order to not block up
            # the system. (Note: The blocking of system is mostly prevalent in Windows and does
            # not happen very often in Linux. However, out of caution 1 CPU is kept free regardless
            # of the system.).
            small_pool = mlp.Pool(NUM_CPU)
            final = small_pool.starmap(helper.cubic_help,
                                       zip(df_chunks, ids_, itertools.repeat(time_jump)))
            small_pool.close()
            small_pool.join()

            # Append the smaller dataframe to process manager list so that result
            # can be finally merged into a larger dataframe.
            return_list.append(pd.concat(final))

        except ValueError:
            raise ValueError

    @staticmethod
    def _kinematic_ip(dataframe: Union[pd.DataFrame, NumTrajDF], time_jump, return_list):
        """
             Method for Kinematic interpolation of a dataframe based on the time jump provided.
             It interpolates the coordinates based on the Datetime of the dataframe.

             WARNING: Do not use this method directly as it will run slower. Instead,
                     use the method interpolate_position() and specify the ip_type as
                     kinematic to perform kinematic interpolation much faster.

             Parameters
             ----------
                 dataframe: Union[pd.DataFrame, NumTrajDF]
                     The dataframe on which interpolation is to be performed
                 time_jump: float
                     The maximum time difference allowed to have between rows
                 return_list: list
                     The list used by the Multiprocessing manager to get the return values

             Returns
             -------
                 pandas.core.dataframe.DataFrame:
                     The dataframe containing the new interpolated points.

         """
        # First, reset the index, extract the Latitude, Longitude, DateTime and Trajectory ID columns
        # and set the DateTime column only as the index. Then, store all the unique Trajectory IDs in
        # a list.
        dataframe = dataframe.reset_index()[
            [const.DateTime, const.TRAJECTORY_ID, const.LAT, const.LONG]].set_index(const.DateTime)

        # Split the smaller dataframe further into smaller chunks containing only 1
        # Trajectory ID per index.
        ids_ = list(dataframe[const.TRAJECTORY_ID].value_counts().keys())
        df_chunks = [dataframe.loc[dataframe[const.TRAJECTORY_ID] == ids_[i]] for i in range(len(ids_))]

        # Here, create as many processes at once as there are number of CPUs available in
        # the system - 1. One CPU is kept free at all times in order to not block up
        # the system. (Note: The blocking of system is mostly prevalent in Windows and does
        # not happen very often in Linux. However, out of caution 1 CPU is kept free regardless
        # of the system.).
        small_pool = mlp.Pool(NUM_CPU)
        final = small_pool.starmap(helper.kinematic_help,
                                   zip(df_chunks, ids_, itertools.repeat(time_jump)))
        small_pool.close()
        small_pool.join()

        # Append the smaller dataframe to process manager list so that result
        # can be finally merged into a larger dataframe.
        return_list.append(pd.concat(final))

    @staticmethod
    def _random_walk_ip(dataframe: Union[pd.DataFrame, NumTrajDF], time_jump, return_list):
        """
             Method for Random walk interpolation of a dataframe based on the time jump provided.
             It interpolates the coordinates based on the Datetime of the dataframe.

             WARNING: Do not use this method directly as it will run slower. Instead,
                     use the method interpolate_position() and specify the ip_type as
                     random-walk to perform random walk interpolation much faster.


             Parameters
             ----------
                 dataframe: Union[pd.DataFrame, NumTrajDF]
                     The dataframe on which interpolation is to be performed
                 time_jump: float
                     The maximum time difference allowed to have between rows
                 return_list: list
                     The list used by the Multiprocessing manager to get the return values

             Returns
             -------
                 pandas.core.dataframe.DataFrame:
                     The dataframe containing the new interpolated points.

         """
        # First, reset the index, extract the Latitude, Longitude, DateTime and Trajectory ID columns
        # and set the DateTime column only as the index. Then, store all the unique Trajectory IDs in
        # a list.
        dataframe = dataframe.reset_index()[
            [const.DateTime, const.TRAJECTORY_ID, const.LAT, const.LONG]].set_index(const.DateTime)

        # Split the smaller dataframe further into smaller chunks containing only 1
        # Trajectory ID per index.
        ids_ = list(dataframe[const.TRAJECTORY_ID].value_counts().keys())
        df_chunks = [dataframe.loc[dataframe[const.TRAJECTORY_ID] == ids_[i]] for i in range(len(ids_))]

        # Here, create as many processes at once as there are number of CPUs available in
        # the system - 1. One CPU is kept free at all times in order to not block up
        # the system. (Note: The blocking of system is mostly prevalent in Windows and does
        # not happen very often in Linux. However, out of caution 1 CPU is kept free regardless
        # of the system.).
        small_pool = mlp.Pool(NUM_CPU)
        final = small_pool.starmap(helper.random_walk_help,
                                   zip(df_chunks, ids_, itertools.repeat(time_jump)))
        small_pool.close()
        small_pool.join()

        # Append the smaller dataframe to process manager list so that result
        # can be finally merged into a larger dataframe.
        return_list.append(pd.concat(final))

    @staticmethod
    def interpolate_at_time(dataframe: Union[NumTrajDF, pandas.DataFrame], time: Text, ip_type: Text = 'linear'):
        pass



"""
    The semantic features module contains several semantic features like
    intersection of trajectories, stop and stay point detection. Moreover,
    features like distance from Point of Interests, water bodies and other
    demographic features related to the trajectory data are calculated. The
    demographic features are extracted with the help of the python osmnx
    library.

    | Authors: Yaksh J Haranwala, Salman Haidri
"""
import itertools
import multiprocessing
import os
from json import JSONDecodeError
from math import ceil
from typing import Union, Text

import geopandas as gpd
import osmnx as ox
import pandas as pd
from shapely.geometry import Polygon

from ptrail.core.TrajectoryDF import PTRAILDataFrame
from ptrail.features.helper_functions import Helpers
from ptrail.utilities.DistanceCalculator import FormulaLog

num = os.cpu_count()
NUM_CPU = ceil((num * 2) / 3)


class SemanticFeatures:
    @staticmethod
    def visited_location(df: PTRAILDataFrame,
                         geo_layers: Union[pd.DataFrame, gpd.GeoDataFrame],
                         visited_location_name: Text,
                         location_column_name: Text):
        """
            Create a column called visited_Location for all the pastures present in the
            dataset.

            Warning
            -------
                While using this method, make sure that the geo_layers parameter dataframe
                that is being passed into the method has Latitude and Longitude columns with
                columns named as 'lat' and 'lon' respectively. If this format is not followed
                then a KeyError will be thrown.

            Note
            ----
                It is to be noted that depending on the size of the dataset and the surrounding
                data passed in, this function will take longer time to execute if either of the
                datasets is very large. It has been parallelized to make it faster, however, it
                can still take a longer time depending on the size of the data being analyzed.

            Parameters
            ----------
                df: PTRAILDataFrame
                    The dataframe containing the dataset.
                geo_layers: Union[pd.DataFrame, gpd.GeoDataFrame]
                    The Dataframe containing the geographical layers near the trajectory data.
                    It is to be noted
                visited_location_name: Text
                    The location for which it is to be checked whether the objected visited it
                    or not.
                location_column_name: Text
                    The name of the column that contains the location to be checked.

            Returns
            -------
                PTRAILDataFrame:
                    The Dataframe containing a new column indicating whether the animal
                    has visited the pasture or not.

            Raises
            ------
                KeyError:
                    The column or the location name does not exist.
        """
        df = df.reset_index()

        # First, filter out the geo_layers dataset to include only the points of the location
        # specified by the user.
        geo_layers = geo_layers.loc[geo_layers[location_column_name] == visited_location_name]
        
        if len(geo_layers) > 0:
            # Now for the trajectory dataset and the geo layers dataset both, convert them to a
            # GeoDataFrame with the geometry of lat-lon for each point.
            df1 = gpd.GeoDataFrame(df,
                                   geometry=gpd.points_from_xy(df["lon"],
                                                               df["lat"]),
                                   crs={"init": "epsg:4326"})
    
            df2 = gpd.GeoDataFrame(geo_layers,
                                   geometry=gpd.points_from_xy(geo_layers["lon"],
                                                               geo_layers["lat"]),
                                   crs={"init": "epsg:4326"})
    
            # Now, using GeoPandas, find where the trajectory points and the geo-layers
            # point intersect.
            intersection = gpd.overlay(df1, df2, how='intersection')
    
            # Now, in the original dataframe, check which points have intersected
            # with the geo-layers dataset and which ones have not.
            merged = pd.merge(df, intersection, how='outer', indicator=True)['_merge']
    
            # Finally, replace the truth value of the points that have intersected to 1
            # and set it to 0 for the points that have not intersected.
            merged = merged.replace('both', 1)
            merged = merged.replace('left_only', 0)
            merged = merged.replace('right_only', 0)
    
            # Assign the resultant column to the original df and drop the unnecessary column
            # of geometry.
            df[f'Visited_{visited_location_name}'] = merged
            df = df.drop(columns='geometry')
    
            # return merged
            return PTRAILDataFrame(df,
                                   latitude='lat',
                                   longitude='lon',
                                   datetime='DateTime',
                                   traj_id='traj_id')
        else:
            raise KeyError(f"The {visited_location_name} or {location_column_name} does not exist in the dataset.")
        
    @staticmethod
    def visited_poi(df: PTRAILDataFrame,
                    surrounding_data: Union[gpd.GeoDataFrame, pd.DataFrame, PTRAILDataFrame],
                    dist_column_label: Text,
                    nearby_threshold: int):
        """
            Given a surrounding data with information about the distance to the nearest POI source
            from a given coordinate, check whether the objects in the given trajectory data have
            visited/crossed those POIs or not

            Warning
            -------
                It is to be noted that for this method to work, the surrounding dataset NEEDS to have a
                column containing distance to the nearest POI. For more info, see the Starkey habitat
                dataset which has the columns like 'DistCWat' and 'DistEWat'.


            Parameters
            ----------
                df: PTRAILDataFrame
                    The dataframe containing the trajectory data.
                surrounding_data: Union[gpd.GeoDataFrame, pd.DataFrame]
                    The surrounding data that needs to contain the information of distance
                    to the nearest water body.
                dist_column_label: Text
                    The name of the column containing the distance information.
                nearby_threshold: int
                    The maximum distance between the POI and the current location of the object
                    within which the object is considered to be crossing/visiting the POI.

            Returns
            -------
                PTRAILDataFrame:
                    The dataframe containing the new column indicating whether the object
                    at that point is near.

        """
        df_chunks = Helpers._df_split_helper(df)
        print(len(df_chunks))

        # Here, create 2/3rds number of processes as there are CPUs in the system. Some CPUs are
        # kept free at all times in order to not block up the system.
        # (Note: The blocking of system is mostly prevalent in Windows and does not happen very often
        # in Linux. However, out of caution some CPUs are kept free regardless of the system.)
        mp_pool = multiprocessing.Pool(NUM_CPU)
        results = mp_pool.starmap(Helpers.visited_poi_helper,
                                  zip(df_chunks,
                                      itertools.repeat(surrounding_data),
                                      itertools.repeat(dist_column_label),
                                      itertools.repeat(nearby_threshold)
                                      )
                                  )

        mp_pool.close()
        mp_pool.join()

        # Concatenate all the smaller dataframes and return the answer.
        results = pd.concat(results)
        return results

    @staticmethod
    def trajectories_inside_polygon(df: PTRAILDataFrame, polygon: Polygon):
        """
            Given a trajectory dataframe and a Polygon, find out all the trajectories
            that are inside the given polygon.

            Warning
            -------
                While creating a polygon, the format of the coordinates is: (longitude, latitude)
                instead of (latitude, longitude). Beware of that, otherwise the results will be
                incorrect.

            Parameters
            ----------
                df: PTRAILDataFrame
                    The dataframe containing the trajectory data.
                polygon: Polygon
                    The polygon inside which the points are to be found.

            Returns
            -------
                PTRAILDataFrame:
                    A dataframe containing trajectories that are inside the polygon.
        """
        # Convert the original dataframe and the polygon to a GeoPandasDataframe.
        df1 = gpd.GeoDataFrame(df.reset_index(),
                               geometry=gpd.points_from_xy(df["lon"],
                                                           df["lat"]),
                               crs={"init": "epsg:4326"})

        df2 = gpd.GeoDataFrame(geometry=gpd.GeoSeries(polygon),
                               crs={"init": "epsg:4326"})

        # Find the points on intersections between the 2 DataFrames.
        intersection = gpd.overlay(df1=df1, df2=df2, how='intersection')
        intersection = intersection.drop(columns=['geometry'])

        # Convert the filtered DF back to PTRAILDataFrame and return it.
        return PTRAILDataFrame(data_set=intersection,
                               datetime='DateTime',
                               traj_id='traj_id',
                               latitude='lat',
                               longitude='lon')

    @staticmethod
    def traj_intersect_inside_polygon(df1: PTRAILDataFrame,
                                      df2: PTRAILDataFrame,
                                      polygon: Polygon):
        """
            Given a df1 and df2 containing trajectory data along with  polygon,
            check whether the trajectory/trajectories are inside the polygon
            and if they are, whether the intersect at any point or not.

            Warning
            -------
                While creating a polygon, the format of the coordinates is: (longitude, latitude)
                instead of (latitude, longitude). Beware of that, otherwise the results will be
                incorrect.

            Note
            ----
                It is to be noted that df1 and df2 should only contain trajectory
                data of only one trajectory each. If they contain more than one
                trajectories, then the results might be unexpected.

            Parameters
            ----------
                df1: PTRAILDataFrame
                    Trajectory Dataframe 1.
                df2: PTRAILDataFrame
                    Trajectory Dataframe 2.
                polygon: Polygon
                    The area inside which it is to be determined if the trajectories
                    intersect or not.

            Returns
            -------
                PTRAILDataFrame:
                    A dataframe containing trajectories that are inside the polygon.
                geopandas.GeoDataFrame:
                    An empty dataframe if both the trajectories do not intersect.
        """
        df1, df2 = df1.reset_index(), df2.reset_index()
        # Convert df1 to GeoDataFrame with correct geometry and the correct CRS.
        gpd_one = gpd.GeoDataFrame(df1,
                                   geometry=gpd.points_from_xy(df1["lon"],
                                                               df1["lat"]),
                                   crs={"init": "epsg:4326"})
        # Convert df2 to GeoDataFrame with correct geometry and the correct CRS.
        gpd_two = gpd.GeoDataFrame(df2,
                                   geometry=gpd.points_from_xy(df2["lon"],
                                                               df2["lat"]),
                                   crs={"init": "epsg:4326"})

        # Convert the polygon to a GeoDataFrame.
        poly_gpd = gpd.GeoDataFrame(geometry=gpd.GeoSeries(polygon),
                                    crs={"init": "epsg:4326"})

        # First, checking intersection between t1 and polygon.
        inter_one = gpd.overlay(gpd_one, poly_gpd, 'intersection')

        # Now, checking intersection between t2 and polygon.
        inter_two = gpd.overlay(gpd_two, poly_gpd, 'intersection')

        # Now, checking intersection between inter t1 and t2.
        final = gpd.overlay(inter_one, inter_two, 'intersection')

        if len(final) > 0:
            to_return = final[['traj_id_1', 'traj_id_2', 'lat_1', 'lon_1', 'DateTime_1', 'DateTime_2']]
            to_return = to_return.rename(columns={'lat_1': 'lat',
                                                  'lon_1': 'lon'})
            cols = ['traj_id_1', 'DateTime_1', 'lat', 'lon', 'traj_id_2', 'DateTime_2']
            to_return = to_return[cols]
            return to_return
        else:
            return final

    @staticmethod
    def nearest_poi(coords: tuple, dist_threshold, tags: dict):
        """
            Given a coordinate point and a distance threshold, find the Point of Interest
            which is the nearest to it within the given distance threshold.


            Warning
            -------
                The users are advised the be mindful about the tags being passed in as
                parameter. More the number of tags, longer will the OSMNx library take
                to download the information from the OpenStreetNetwork maps. Moreover,
                an active internet connection is also required to execute this function.

            Note
            ----
                If several tags (POIs) are given in, then the method will find the closest
                one based on the distance and return it and will not given out the others
                that may or may not be present within the threshold of the given point.

            Parameters
            ----------
                coords: tuple
                    The point near which the bank is to be found.
                dist_threshold:
                    The maximum distance from the point within which
                    the distance is to be calculated.
                tags: dict
                    The dictionary containing tags of Points of interest.

            Returns
            -------
                pandas.core.dataframe.DataFrame:
                    A pandas DF containing the info about the nearest bank from
                    the given point.

            Raises
            ------
                JSONDecodeError:
                    One or more given tags are invalid.
        """
        try:
            # Given the tag, the point and the distance threshold, use the osmnx
            # library the fetch the nearby POIs.
            poi = ox.geometries_from_point(center_point=coords,
                                           dist=dist_threshold,
                                           tags=tags)

            # Check whether there are any errors in the tag or if there are actually
            # no point of interests with the given tags nearby the given point.
            if len(poi) > 0:
                # Remove unnecessary column of element_type.
                poi = poi.reset_index().loc[poi.reset_index()['element_type'] == 'node']

                # Convert the Geometry to lat and lon.
                lat = list(poi['geometry'].apply(lambda p: p.y))
                lon = list(poi['geometry'].apply(lambda p: p.x))

                # Check the distances from all nearby resultant POIs and then find out the
                # one which is the nearest to the given point. Mind that the distance
                # calculated is the haversine distance and not walking/driving distance.
                dists = []
                for i in range(len(lat)):
                    dists.append(FormulaLog.haversine_distance(coords[0], coords[1], lat[i], lon[i]))

                poi[f'Distance_from_{coords}'] = dists

                # Return the nearest POI.
                return poi.loc[poi[f'Distance_from_{coords}'] ==
                               poi[f'Distance_from_{coords}'].min()].reset_index().drop(columns=['element_type',
                                                                                                 'index'])
            else:
                # If there are no POIs near the given point, return an empty list.
                return []

        except JSONDecodeError:
            raise ValueError("The tags provided are invalid. Please check your tags and try again.")

    # @staticmethod
    # def detect_herd(df: PTRAILDataFrame,
    #                 surroundings: Union[pd.DataFrame, gpd.GeoDataFrame],
    #                 time_threshold: int):
    #     """
    #         Given a Trajectory Dataframe and another dataframe containing surrounding data,
    #         detect a herd behaviour in a particular region of the surroundings.
    #
    #         Note
    #         ----
    #             The time_threshold is given in seconds.
    #
    #         Parameters
    #         ----------
    #             df: PTRAILDataFrame
    #                 The dataframe containing Trajectory Data.
    #             surroundings: Union[pd.DataFrame, gpd.GeoDataFrame]
    #                 The dataframe containing surrounding data.
    #             time_threshold: int
    #                 The maximum time between 2 points of intersection lesser than which
    #                 the objects are considered to be in herd.
    #     """
    #     # coords = list(zip(surroundings['lon'], (surroundings['lat'])))
    #     # poly = Polygon(coords)
    #     #
    #     # # First, check which of the trajectories are inside the polygon.
    #     # inside_polygon = SemanticFeatures.trajectories_inside_polygon(df, poly)
    #     #
    #     # return inside_polygon
    #
    #     # First split the dataframes based on set of Trajectory ids.
    #     # As of now, each smaller chunk is supposed to have data of 100
    #     # trajectory IDs max
    #     ids_ = list(df.traj_id.value_counts().keys())
    #
    #     df_chunks = [df.loc[df.index.get_level_values(const.TRAJECTORY_ID) == ids_[i]]
    #                  for i in range(len(ids_))]
    #
    #     main = []
    #     for i in range(len(df_chunks) - 1):
    #         df1 = df_chunks[i]
    #         df2 = df_chunks[i + 1]
    #         intersect = SemanticFeatures.traj_intersect_inside_polygon(df1, df2, poly)
    #
    #         time_del = (np.abs((intersect['DateTime_2'] - intersect['DateTime_1'])).dt.total_seconds()).to_numpy()
    #
    #         indices = np.argwhere(time_del <= time_threshold).flatten()
    #         print(indices)
    #
    #         filt_df = intersect.iloc[indices]
    #         main.append(filt_df)
    #
    #     return main



"""
    The spatial_features module contains several functions of the library
    that calculates kinematic features based on the coordinates of points
    provided in the data. This module mostly extracts and modifies data
    collected from some existing dataframe and appends these information
    to them. Inspiration of lots of functions in this module is taken from
    the PyMove library.


    | Authors: Yaksh J Haranwala, Salman Haidri

    References
    ----------
            Arina De Jesus Amador Monteiro Sanches. “Uma Arquitetura E Imple-menta ̧c ̃ao Do M ́odulo De
            Pr ́e-processamento Para Biblioteca Pymove”.Bachelor’s thesis. Universidade Federal Do Cear ́a, 2019
"""
import itertools
import multiprocessing
import os
from math import ceil
from typing import Optional, Text

import numpy as np
import pandas as pd

from ptrail.core.TrajectoryDF import PTRAILDataFrame
from ptrail.features.helper_functions import Helpers as helpers
from ptrail.utilities import constants as const
from ptrail.utilities.DistanceCalculator import FormulaLog as calc
from ptrail.utilities.exceptions import *

num = os.cpu_count()
NUM_CPU = ceil((num * 2) / 3)


class KinematicFeatures:
    @staticmethod
    def get_bounding_box(dataframe: PTRAILDataFrame):
        """
            Return the bounding box of the Trajectory data. Essentially, the bounding box is of
            the following format:
                (min Latitude, min Longitude, max Latitude, max Longitude).

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe containing the trajectory data.

            Returns
            -------
                tuple:
                    The bounding box of the trajectory
        """
        return (
            dataframe[const.LAT].min(),
            dataframe[const.LONG].min(),
            dataframe[const.LAT].max(),
            dataframe[const.LONG].max(),
        )

    @staticmethod
    def get_start_location(dataframe: PTRAILDataFrame, traj_id=None):
        """
            Get the starting location of an object's trajectory in the data.

            Note
            ----
                If the user does not give in any traj_id, then the library,
                by default gives out the start locations of all the unique
                trajectory ids present in the data.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The PTRAILDataFrame storing the trajectory data.
                traj_id:
                    The ID of the object whose start location is to be found.

            Returns
            -------
                tuple:
                    The (lat, longitude) tuple containing the start location.
                pandas.core.dataframe.DataFrame:
                    The dataframe containing start locations of all trajectory IDs.
        """
        # If traj_id is None, find the start times of all the unique trajectories present in the data.
        # Else first filter out a dataframe containing the given traj_id and then return the start
        # location of that point.
        dataframe = dataframe.copy().reset_index()
        if traj_id is None:
            ids_ = dataframe[const.TRAJECTORY_ID].value_counts(ascending=True).keys().to_list()
            # Get the ideal number of IDs by which the dataframe is to be split.
            split_factor = helpers._get_partition_size(len(ids_))
            ids_ = [ids_[i: i + split_factor] for i in range(0, len(ids_), split_factor)]

            # Here, create 2/3rds number of processes as there are in the system. Some CPUs are
            # kept free at all times in order to not block up the system.
            # (Note: The blocking of system is mostly prevalent in Windows and does not happen very often
            # in Linux. However, out of caution some CPUs are kept free regardless of the system.)
            mp_pool = multiprocessing.Pool(NUM_CPU)
            results = mp_pool.starmap(helpers.start_location_helper, zip(itertools.repeat(dataframe), ids_))
            mp_pool.close()
            mp_pool.join()

            # Concatenate all the smaller dataframes and return the answer.
            results = pd.concat(results)
            return results

        else:
            filt = (dataframe.loc[dataframe[const.TRAJECTORY_ID] == traj_id, [const.DateTime, const.LAT, const.LONG]])
            start_loc = (filt.loc[filt[const.DateTime] == filt[const.DateTime].min(),
                                  [const.LAT, const.LONG]]).reset_index()
            if len(start_loc) == 0:
                return f"Trajectory ID: {traj_id} does not exist in the dataset. Please try again!"
            else:
                return start_loc[const.LAT][0], start_loc[const.LONG][0]

    @staticmethod
    def get_end_location(dataframe: PTRAILDataFrame, traj_id: Optional[Text] = None):
        """
            Get the ending location of an object's trajectory in the data.

            Note
            ----
                If the user does not give in any traj_id, then the library,
                by default gives out the end locations of all the unique
                trajectory ids present in the data.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The PTRAILDataFrame storing the trajectory data.
                traj_id
                    The ID of the trajectory whose end location is to be found.

            Returns
            -------
                tuple:
                    The (lat, longitude) tuple containing the end location.
                pandas.core.dataframe.DataFrame:
                    The dataframe containing start locations of all trajectory IDs.
        """
        # If traj_id is None, find the end times of all the unique trajectories present in the data.
        # Else first filter out a dataframe containing the given traj_id and then return the end
        # location of that point.
        dataframe = dataframe.copy().reset_index()
        if traj_id is None:
            ids_ = dataframe[const.TRAJECTORY_ID].value_counts(ascending=True).keys().to_list()
            # Get the ideal number of IDs by which the dataframe is to be split.
            split_factor = helpers._get_partition_size(len(ids_))
            ids_ = [ids_[i: i + split_factor] for i in range(0, len(ids_), split_factor)]

            # Here, create 2/3rds number of processes as there are in the system. Some CPUs are
            # kept free at all times in order to not block up the system.
            # (Note: The blocking of system is mostly prevalent in Windows and does not happen very often
            # in Linux. However, out of caution some CPUs are kept free regardless of the system.)
            mp_pool = multiprocessing.Pool(NUM_CPU)
            results = mp_pool.starmap(helpers.end_location_helper, zip(itertools.repeat(dataframe), ids_))
            mp_pool.close()
            mp_pool.join()

            # Concatenate all the smaller dataframes and return the answer.
            results = pd.concat(results)
            return results
        else:
            filt = (dataframe.loc[dataframe[const.TRAJECTORY_ID] == traj_id, [const.DateTime, const.LAT, const.LONG]])
            start_loc = (filt.loc[filt[const.DateTime] == filt[const.DateTime].max(),
                                  [const.LAT, const.LONG]]).reset_index()
            if len(start_loc) == 0:
                return f"Trajectory ID: {traj_id} does not exist in the dataset. Please try again!"
            else:
                return start_loc[const.LAT][0], start_loc[const.LONG][0]

    @staticmethod
    def create_distance_column(dataframe: PTRAILDataFrame):
        """
            Create a column called Dist_prev_to_curr containing distance between 2 consecutive points.
            The distance calculated is the Great-Circle (Haversine) distance.

            Note
            ----
                When the trajectory ID changes in the data, then the distance calculation again starts
                from the first point of the new trajectory ID and the distance-value of the first point
                of the new Trajectory ID will be set to 0.

            Note
            ----
                The Distance calculated here is the distance between 2 consecutive points of the same
                trajectory. Furthermore, the distance yielded is in metres (m).

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The data where distance is to be calculated.

            Returns
            -------
                PTRAILDataFrame:
                    The dataframe containing the resultant Distance_prev_to_curr column.
        """
        # Case-1: The number of unique Trajectory IDs is less than 100.
        if dataframe.traj_id.nunique() < const.MIN_IDS:
            result = helpers.distance_between_consecutive_helper(dataframe)
            return PTRAILDataFrame(result, const.LAT, const.LONG,
                                   const.DateTime, const.TRAJECTORY_ID)
        # Case-2: The number of unique Trajectory IDs is significant.
        else:
            # splitting the dataframe according to trajectory ids.
            df_chunks = helpers._df_split_helper(dataframe)

            # Here, create 2/3rds number of processes as there are in the system. Some CPUs are
            # kept free at all times in order to not block up the system.
            # (Note: The blocking of system is mostly prevalent in Windows and does not happen very often
            # in Linux. However, out of caution some CPUs are kept free regardless of the system.)
            multi_pool = multiprocessing.Pool(NUM_CPU)
            result = multi_pool.map(helpers.distance_between_consecutive_helper, df_chunks)
            multi_pool.close()
            multi_pool.join()

            # merge the smaller pieces and then return the dataframe converted to PTRAILDataFrame.
            return PTRAILDataFrame(pd.concat(result), const.LAT, const.LONG,
                                   const.DateTime, const.TRAJECTORY_ID)

    @staticmethod
    def create_distance_from_start_column(dataframe: PTRAILDataFrame):
        """
            Create a column containing distance between the start location and the rest of the
            points using Haversine formula. The distance calculated is the Great-Circle distance.

            Note
            ----
                When the trajectory ID changes in the data, then the distance calculation again
                starts from the first point of the new trajectory ID and the first distance of the
                new trajectory ID will be set to 0.

            Note
            ----
                The Distance calculated here is the distance between the start point and the current points of the same
                trajectory. Furthermore, the distance yielded is in metres (m).

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The data where distance is to be calculated.

            Returns
            -------
                PTRAILDataFrame:
                    The dataframe containing the resultant Distance_start_to_curr column.
        """
        # Case-1: The number of unique Trajectory IDs is less than 100.
        if dataframe.traj_id.nunique() < const.MIN_IDS:
            result = helpers.distance_from_start_helper(dataframe)
            return PTRAILDataFrame(result, const.LAT, const.LONG,
                                   const.DateTime, const.TRAJECTORY_ID)

        # Case-2: The number of unique Trajectory IDs is significant.
        else:
            # splitting the dataframe according to trajectory ids.
            df_chunks = helpers._df_split_helper(dataframe)

            # Here, create 2/3rds number of processes as there are in the system. Some CPUs are
            # kept free at all times in order to not block up the system.
            # (Note: The blocking of system is mostly prevalent in Windows and does not happen very often
            # in Linux. However, out of caution some CPUs are kept free regardless of the system.)
            multi_pool = multiprocessing.Pool(NUM_CPU)
            result = multi_pool.map(helpers.distance_from_start_helper, df_chunks)
            multi_pool.close()
            multi_pool.join()

            # merge the smaller pieces and then return the dataframe converted to PTRAILDataFrame.
            return PTRAILDataFrame(pd.concat(result), const.LAT, const.LONG,
                                   const.DateTime, const.TRAJECTORY_ID)

    @staticmethod
    def distance_travelled_by_date_and_traj_id(dataframe: PTRAILDataFrame, date, traj_id):
        """
            Given a date and trajectory ID, calculate the total distance
            covered in the trajectory on that particular date.

            Note
            ----
                The distance yielded is in metres (m).

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe in which teh actual data is stored.
                date: Text
                    The Date on which the distance covered is to be calculated.
                traj_id: Text
                    The trajectory ID for which the distance covered is to be calculated.

            Returns
            -------
                float:
                    The total distance covered on that date by that trajectory ID.

            Raises
            ------
                KeyError:
                    Traj_id is not present in the arguments passed.
        """
        # First, reset the index of the dataframe.
        # Then, filter the dataframe based on Date and Trajectory ID if given by user.
        dataframe = dataframe.reset_index()
        filt = dataframe.loc[dataframe[const.DateTime].dt.date == pd.to_datetime(date)]
        small = filt.loc[filt[const.TRAJECTORY_ID] == traj_id]

        if len(small) > 0:
            # First, lets fetch the latitude and longitude columns from the dataset and store it
            # in a numpy array.
            traj_ids = np.array(small.reset_index()[const.TRAJECTORY_ID])
            latitudes = np.array(small[const.LAT])
            longitudes = np.array(small[const.LONG])
            distances = np.zeros(len(traj_ids))

            # Now, lets calculate the Great-Circle (Haversine) distance between the 2 points and store
            # each of the values in the distance numpy array.
            for i in range(len(latitudes) - 1):
                distances[i + 1] = calc.haversine_distance(latitudes[i], longitudes[i], latitudes[i + 1], longitudes[i + 1])

            return np.sum(distances)  # Sum all the distances and return the total path length.
        else:
            raise KeyError(f"The Trajectory ID: {traj_id} is not present in the Dataset.")

    @staticmethod
    def create_point_within_range_column(dataframe: PTRAILDataFrame, coordinates: tuple,
                                         dist_range: float):
        """
            Check how many points are within the range of the given coordinate by first making a column
            containing the distance between the given coordinate and rest of the points in dataframe by calling
            create_distance_from_point() and then comparing each point using the condition if it's within the
            range and appending the values in a column and attaching it to the dataframe.

            Note
            ----
                The dist_range parameter is given in metres.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe on which the point within range calculation is to be done.
                coordinates: tuple
                    The coordinates from which the distance is to be calculated.
                dist_range: float
                    The range within which the resultant distance from the coordinates should lie.

            Returns
            -------
                PTRAILDataFrame:
                    The dataframe containing the resultant Within_x_m_from_(x,y) column.
        """
        # splitting the dataframe according to trajectory ids
        df_chunks = helpers._df_split_helper(dataframe)

        # Here, create 2/3rds number of processes as there are in the system. Some CPUs are
        # kept free at all times in order to not block up the system.
        # (Note: The blocking of system is mostly prevalent in Windows and does not happen very often
        # in Linux. However, out of caution some CPUs are kept free regardless of the system.)
        pool = multiprocessing.Pool(NUM_CPU)
        args = zip(df_chunks, itertools.repeat(coordinates), itertools.repeat(dist_range))
        result = pool.starmap(helpers.point_within_range_helper, args)
        pool.close()
        pool.join()

        # Now lets join all the smaller partitions and return the resultant dataframe
        result = pd.concat(result)
        return PTRAILDataFrame(result.reset_index(), const.LAT, const.LONG, const.DateTime, const.TRAJECTORY_ID)

    @staticmethod
    def create_distance_from_point_column(dataframe: PTRAILDataFrame, coordinates: tuple):
        """
            Given a point, this function calculates the distance between that point and all the
            points present in the dataframe and adds that column into the dataframe.

            Note
            ----
                The distance yielded here is in metres.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe on which calculation is to be done.
                coordinates: tuple
                    The coordinates from which the distance is to be calculated.

            Returns
            -------
                PTRAILDataFrame:
                    The dataframe containing the resultant Distance_from_(x, y) column.
        """
        # dataframe = dataframe.reset_index()
        # splitting the dataframe according to trajectory ids
        df_chunks = helpers._df_split_helper(dataframe)

        # Here, create 2/3rds number of processes as there are in the system. Some CPUs are
        # kept free at all times in order to not block up the system.
        # (Note: The blocking of system is mostly prevalent in Windows and does not happen very often
        # in Linux. However, out of caution some CPUs are kept free regardless of the system.)
        pool = multiprocessing.Pool(NUM_CPU)
        answer = pool.starmap(helpers.distance_from_given_point_helper, zip(df_chunks, itertools.repeat(coordinates)))
        pool.close()
        pool.join()

        # Now lets join all the smaller partitions and then add the Distance to the
        # specific point column.
        answer = pd.concat(answer)

        # return the answer dataframe converted to PTRAILDataFrame.
        return PTRAILDataFrame(answer.reset_index(), const.LAT, const.LONG, const.DateTime, const.TRAJECTORY_ID)

    @staticmethod
    def create_speed_column(dataframe: PTRAILDataFrame):
        """
            Create a column containing speed of the object from the previous point
            to the current point.

            Note
            ----
                When the trajectory ID changes in the data, then the speed calculation again
                starts from the first point of the new trajectory ID and the speed of the
                first point of the new trajectory ID will be set to 0.

            Note
            ----
                The Speed calculated here is the speed between 2 consecutive points of the same
                trajectory. Furthermore, the speed yielded is in metres/second (m/s).

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe on which the calculation of speed is to be done.

            Returns
            -------
                PTRAILDataFrame:
                    The dataframe containing the resultant Speed_prev_to_curr column.
        """
        # Here, we are using try and catch blocks to check whether the DataFrame has the
        # Distance_prev_to_curr column.
        try:
            # If the Distance_prev_to_curr column is already present in the dataframe,
            # then extract it, calculate the time differences between the consecutive
            # rows in the dataframe and then calculate distances/time_deltas in order to
            # calculate the speed.
            # WARNING!!!! Use dt.total_seconds() as dt.seconds gives false values and as it
            #             does not account for time difference when it is negative.
            distances = dataframe.reset_index()['Distance']
            time_deltas = dataframe.reset_index()[const.DateTime].diff().dt.total_seconds()

            # Assign the new column and return the NumPandasTrajDF.
            dataframe['Speed'] = (distances / time_deltas.dropna()).to_numpy()
            dataframe = dataframe.replace([np.inf, -np.inf], np.nan)
            return PTRAILDataFrame(data_set=dataframe.reset_index(),
                                   datetime='DateTime',
                                   traj_id='traj_id',
                                   latitude='lat',
                                   longitude='lon')

        except KeyError:
            # If the Distance_prev_to_curr column is not present in the Dataframe and a KeyError
            # is thrown, then catch it and the overridden behaviour is as follows:
            #   1. Calculate the distance by calling the create_distance_column() function.
            #   2. Calculate the time deltas.
            #   3. Divide the 2 values to calculate the speed.
            # WARNING!!!! Use dt.total_seconds() as dt.seconds gives false values and as it
            #             does not account for time difference when it is negative.
            dataframe = KinematicFeatures.create_distance_column(dataframe)
            distances = dataframe.reset_index()['Distance']
            time_deltas = dataframe.reset_index()[const.DateTime].diff().dt.total_seconds()

            # Assign the column and return the NumPandasTrajDF.
            dataframe['Speed'] = (distances / time_deltas).to_numpy(dtype=np.float64)
            dataframe = dataframe.replace([np.inf, -np.inf], np.nan)
            return PTRAILDataFrame(data_set=dataframe.reset_index(),
                                   datetime='DateTime',
                                   traj_id='traj_id',
                                   latitude='lat',
                                   longitude='lon')

    @staticmethod
    def create_acceleration_column(dataframe: PTRAILDataFrame):
        """
            Create a column containing acceleration of the object from the previous to the current
            point.

            Note
            ----
                The acceleration calculated here is the acceleration between 2 consecutive points of the same
                trajectory. Furthermore, the acceleration yielded is in metres/second^2 (m/s^2).

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe on which the calculation of acceleration is to be done.

            Returns
            -------
                PTRAILDataFrame:
                    The dataframe containing the resultant Acceleration_prev_to_curr column.
        """
        # Try catch is used to check if speed column is present or not
        try:
            # When Speed column is present extract the data from there and then take calculate the time delta
            # And use that to calculate acceleration by dividing speed by time delta and then add the column to
            # the dataframe
            # WARNING!!!! Use dt.total_seconds() as dt.seconds gives false values and as it
            #             does not account for time difference when it is negative.
            speed_deltas = dataframe.reset_index()['Speed'].diff()
            time_deltas = dataframe.reset_index()[const.DateTime].diff().dt.total_seconds()

            dataframe['Acceleration'] = (speed_deltas / time_deltas).to_numpy()
            dataframe = dataframe.replace([np.inf, -np.inf], np.nan)
            return PTRAILDataFrame(data_set=dataframe.reset_index(),
                                   datetime='DateTime',
                                   traj_id='traj_id',
                                   latitude='lat',
                                   longitude='lon')

        except KeyError:
            # When Speed column is not present then first call create_speed_from_prev_column() function to make
            # the speed column and then follow the steps mentioned above
            # WARNING!!!! Use dt.total_seconds() as dt.seconds gives false values and as it
            #             does not account for time difference when it is negative.
            dataframe = KinematicFeatures.create_speed_column(dataframe)
            speed_deltas = dataframe.reset_index()['Speed'].diff()
            time_deltas = dataframe.reset_index()[const.DateTime].diff().dt.total_seconds()

            dataframe['Acceleration'] = (speed_deltas / time_deltas).to_numpy()
            dataframe = dataframe.replace([np.inf, -np.inf], np.nan)
            return PTRAILDataFrame(data_set=dataframe.reset_index(),
                                   datetime='DateTime',
                                   traj_id='traj_id',
                                   latitude='lat',
                                   longitude='lon')

    @staticmethod
    def create_jerk_column(dataframe: PTRAILDataFrame):
        """
            Create a column containing jerk of the object from previous to the current
            point.

            Note
            ----
                The jerk calculated here is the jerk between 2 consecutive points of the same
                trajectory. Furthermore, the jerk yielded is in metres/second^3 (m/s^3).

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe on which the calculation of jerk is to be done.

            Returns
            -------
                PTRAILDataFrame:
                    The dataframe containing the resultant jerk_prev_to_curr column.
        """
        # Try catch is used to check if acceleration column is present or not
        try:
            # When acceleration column is present extract the data from there and then take calculate the time delta
            # And use that to calculate acceleration by dividing speed_delta by time delta and then add the column to
            # the dataframe
            # WARNING!!!! Use dt.total_seconds() as dt.seconds gives false values and as it
            #             does not account for time difference when it is negative.
            acceleration_deltas = dataframe.reset_index()['Acceleration'].diff()
            time_deltas = dataframe.reset_index()[const.DateTime].diff().dt.total_seconds()

            dataframe['Jerk'] = (acceleration_deltas / time_deltas).to_numpy()
            dataframe = dataframe.replace([np.inf, -np.inf], np.nan)
            return PTRAILDataFrame(data_set=dataframe.reset_index(),
                                   datetime='DateTime',
                                   traj_id='traj_id',
                                   latitude='lat',
                                   longitude='lon')

        except KeyError:
            # When Speed column is not present then first call create_speed_from_prev_column() function to make
            # the speed column and then follow the steps mentioned above
            # WARNING!!!! Use dt.total_seconds() as dt.seconds gives false values and as it
            #             does not account for time difference when it is negative.
            dataframe = KinematicFeatures.create_acceleration_column(dataframe)
            acceleration_deltas = dataframe.reset_index()['Acceleration'].diff()
            time_deltas = dataframe.reset_index()[const.DateTime].diff().dt.total_seconds()

            dataframe['Jerk'] = (acceleration_deltas / time_deltas).to_numpy()
            dataframe = dataframe.replace([np.inf, -np.inf], np.nan)
            return PTRAILDataFrame(data_set=dataframe.reset_index(),
                                   datetime='DateTime',
                                   traj_id='traj_id',
                                   latitude='lat',
                                   longitude='lon')

    @staticmethod
    def create_bearing_column(dataframe: PTRAILDataFrame):
        """
            Create a column containing bearing between 2 consecutive points. Bearing is also
            referred as "Forward Azimuth" sometimes. Bearing/Forward Azimuth is defined as
            follows:
                Bearing is the horizontal angle between the direction of an object and another
                object, or between the object and the True North.

            Note
            ----
                The bearing calculated here is the bearing between 2 consecutive points of the same
                trajectory. Furthermore, the bearing yielded is in degrees.


            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe on which the bearing is to be calculated.

            Returns
            -------
                PTRAILDataFrame:
                        The dataframe containing the resultant Bearing_from_prev column.
        """
        # if cpu_count <= 1:
        #     cpu_count = 1
        # elif cpu_count >= NUM_CPU:
        #     cpu_count = NUM_CPU - 1
        # else:
        #     cpu_count = cpu_count

        # Case-1: The number of unique Trajectory IDs is less than x.
        if dataframe.traj_id.nunique() < const.MIN_IDS:
            result = helpers.bearing_helper(dataframe)
            return PTRAILDataFrame(result, const.LAT, const.LONG,
                                   const.DateTime, const.TRAJECTORY_ID)

        # Case-2: The number unique Trajectory IDs is significant.
        else:
            # splitting the dataframe according to trajectory ids.
            df_chunks = helpers._df_split_helper(dataframe)

            # Here, create 2/3rds number of processes as there are in the system. Some CPUs are
            # kept free at all times in order to not block up the system.
            # (Note: The blocking of system is mostly prevalent in Windows and does not happen very often
            # in Linux. However, out of caution some CPUs are kept free regardless of the system.)
            multi_pool = multiprocessing.Pool(NUM_CPU)
            result = multi_pool.map(helpers.bearing_helper, df_chunks)
            multi_pool.close()
            multi_pool.join()

            # merge the smaller pieces and then return the dataframe converted to PTRAILDataFrame.
            dataframe = pd.concat(result)
            dataframe = dataframe.replace([np.inf, -np.inf], np.nan)
            return PTRAILDataFrame(dataframe, const.LAT, const.LONG,
                                   const.DateTime, const.TRAJECTORY_ID)

    @staticmethod
    def create_bearing_rate_column(dataframe: PTRAILDataFrame):
        """
            Calculates the bearing rate of the consecutive points. And adding that column into
            the dataframe

            Note
            ----
                The bearing calculated here is the bearing between 2 consecutive points of the same
                trajectory. Furthermore, the bearing yielded is in degrees/second.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe on which the bearing rate is to be calculated

            Returns
            -------
                PTRAILDataFrame:
                    The dataframe containing the resultant Bearing_rate_from_prev column.
        """
        # Try catch to check for Bearing column
        try:
            # If Bearing from previous column is present, extract that and then calculate time_deltas
            # Using these calculate Bearing_rate_from_prev by dividing bearing_deltas with time_deltas
            # And then adding the column to the dataframe
            # WARNING!!!! Use dt.total_seconds() as dt.seconds gives false values and as it
            #             does not account for time difference when it is negative.
            bearing_deltas = dataframe.reset_index()['Bearing'].diff()
            time_deltas = dataframe.reset_index()[const.DateTime].diff().dt.total_seconds()

            dataframe['Bearing_Rate'] = (bearing_deltas / time_deltas).to_numpy()
            dataframe = dataframe.replace([np.inf, -np.inf], np.nan)
            return PTRAILDataFrame(data_set=dataframe.reset_index(),
                                   datetime='DateTime',
                                   traj_id='traj_id',
                                   latitude='lat',
                                   longitude='lon')
        except KeyError:
            # Similar to the step above but just makes the Bearing column first
            # WARNING!!!! Use dt.total_seconds() as dt.seconds gives false values and as it
            #             does not account for time difference when it is negative.
            dataframe = KinematicFeatures.create_bearing_column(dataframe)
            bearing_deltas = dataframe.reset_index()['Bearing'].diff()
            time_deltas = dataframe.reset_index()[const.DateTime].diff().dt.total_seconds()

            dataframe['Bearing_Rate'] = (bearing_deltas / time_deltas).to_numpy()
            dataframe = dataframe.replace([np.inf, -np.inf], np.nan)
            return PTRAILDataFrame(data_set=dataframe.reset_index(),
                                   datetime='DateTime',
                                   traj_id='traj_id',
                                   latitude='lat',
                                   longitude='lon')

    @staticmethod
    def create_rate_of_br_column(dataframe: PTRAILDataFrame):
        """
            Calculates the rate of bearing rate of the consecutive points.
            And then adding that column into the dataframe.

            Note
            ----
                The rate of bearing rate calculated here is the rate of bearing rate between 2
                consecutive points of the same trajectory.
                Furthermore, the bearing yielded is in degrees.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe on which the rate of bearing rate is to be calculated

            Returns
            -------
                PTRAILDataFrame:
                    The dataframe containing the resultant Rate_of_bearing_rate_from_prev column
        """
        # Try catch to check for Bearing Rate column
        try:
            # If Bearing from previous column is present, extract that and then calculate time_deltas
            # Using these calculate Bearing_rate_from_prev by dividing bearing_deltas with time_deltas
            # And then adding the column to the dataframe
            # WARNING!!!! Use dt.total_seconds() as dt.seconds gives false values and as it
            #             does not account for time difference when it is negative.
            bearing_rate_deltas = dataframe.reset_index()['Bearing_Rate'].diff()
            time_deltas = dataframe.reset_index()[const.DateTime].diff().dt.total_seconds()

            dataframe['Rate_of_bearing_rate'] = (bearing_rate_deltas / time_deltas).to_numpy()
            dataframe = dataframe.replace([np.inf, -np.inf], np.nan)
            return PTRAILDataFrame(data_set=dataframe.reset_index(),
                                   datetime='DateTime',
                                   traj_id='traj_id',
                                   latitude='lat',
                                   longitude='lon')
        except KeyError:
            # Similar to the step above but just makes the Bearing column first
            # WARNING!!!! Use dt.total_seconds() as dt.seconds gives false values and as it
            #             does not account for time difference when it is negative.
            dataframe = KinematicFeatures.create_bearing_rate_column(dataframe)
            bearing_rate_deltas = dataframe.reset_index()['Bearing'].diff()
            time_deltas = dataframe.reset_index()[const.DateTime].diff().dt.total_seconds()

            dataframe['Rate_of_bearing_rate'] = (bearing_rate_deltas / time_deltas).to_numpy()
            dataframe = dataframe.replace([np.inf, -np.inf], np.nan)
            return PTRAILDataFrame(data_set=dataframe.reset_index(),
                                   datetime='DateTime',
                                   traj_id='traj_id',
                                   latitude='lat',
                                   longitude='lon')

    @staticmethod
    def get_distance_travelled_by_traj_id(dataframe: PTRAILDataFrame, traj_id: Text):
        """
            Given a trajectory ID, calculate the total distance covered by the trajectory.
            NOTE: The distance calculated is in metres.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe containing the entire dataset.
                traj_id: Text
                    The trajectory ID for which the distance covered is to be calculated.

            Returns
            -------
                float:
                    The distance covered by the trajectory

            Raises
            ------
                MissingTrajIDException:
                    The Trajectory ID given by the user is not present in the dataset.
        """
        # First, filter the dataframe and create a smaller dataframe containing only the
        # trajectory points of the specified trajectory ID.
        dataframe = dataframe.reset_index()
        filtered_df = dataframe.loc[dataframe[const.TRAJECTORY_ID] == traj_id]

        # Now, check if the smaller dataframe actually has anything because if the user-given
        # trajectory ID is not present in the dataset, then an empty dataframe is returned and
        # if so raise an exception.
        if len(filtered_df) > 0:
            # First, calculate the distance by calling the distance_between_consecutive_column() function
            # and convert it into a numpy array and then sum the array using nansum() to make sure that
            # NaN values are considered as zeros.
            distances = KinematicFeatures.create_distance_column(filtered_df)[
                'Distance'].to_numpy()
            return np.nansum(distances)
        else:
            raise MissingTrajIDException(f"The Trajectory ID '{traj_id}' is not present in the data."
                                         f"Please check the Trajectory ID and try again.")

    @staticmethod
    def get_number_of_locations(dataframe: PTRAILDataFrame, traj_id: Text = None):
        """
            Get the number of unique coordinates in the dataframe specific to a trajectory ID.

            Note
            ----
                If no Trajectory ID is specified, then the number of unique locations in the
                visited by each trajectory in the dataset is calculated.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe of which the number of locations are to be computed
                traj_id: Text
                    The trajectory id for which the number of unique locations are to be found

            Returns
            -------
                int:
                    The number of unique locations in the dataframe/trajectory id.
                pandas.core.dataframe.DataFrame:
                    The dataframe containing start locations of all trajectory IDs.
        """
        dataframe = dataframe.reset_index()
        if traj_id is None:
            ids_ = dataframe[const.TRAJECTORY_ID].value_counts(ascending=True).keys().to_list()
            # Get the ideal number of IDs by which the dataframe is to be split.
            split_factor = helpers._get_partition_size(len(ids_))
            ids_ = [ids_[i: i + split_factor] for i in range(0, len(ids_), split_factor)]

            # Here, create 2/3rds number of processes as there are in the system. Some CPUs are
            # kept free at all times in order to not block up the system.
            # (Note: The blocking of system is mostly prevalent in Windows and does not happen very often
            # in Linux. However, out of caution some CPUs are kept free regardless of the system.)
            mp_pool = multiprocessing.Pool(NUM_CPU)
            results = mp_pool.starmap(helpers.number_of_location_helper, zip(itertools.repeat(dataframe), ids_))
            mp_pool.close()
            mp_pool.join()

            # Concatenate all the smaller dataframes and return the answer.
            results = pd.concat(results)
            return results

        else:
            filtered_df = dataframe.loc[dataframe[const.TRAJECTORY_ID] == traj_id]
            return filtered_df.groupby([const.LAT, const.LONG]).ngroups

    @staticmethod
    def generate_kinematic_features(dataframe: PTRAILDataFrame):
        """
            Generate all the Kinematic features with a single call of this function.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe on which the features are to be generated.

            Returns
            -------
                PTRAILDataFrame:
                    The dataframe enriched with Kinematic Features.
        """
        to_return = KinematicFeatures.create_distance_column(dataframe)
        to_return = KinematicFeatures.create_distance_from_start_column(to_return)
        to_return = KinematicFeatures.create_jerk_column(to_return)
        to_return = KinematicFeatures.create_rate_of_br_column(to_return)

        return to_return

    @staticmethod
    def generate_kinematic_stats(dataframe: PTRAILDataFrame):
        """
            Generate the statistics of kinematic features for each unique trajectory in
            the dataframe.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe containing the trajectory data.

            Returns
            -------
                pandas.core.dataframe.DataFrame:
                    A pandas dataframe containing stats for all kinematic features for
                    each unique trajectory in the dataframe.
        """
        # Generate kinematic features on the entire dataframe.
        ptdf = KinematicFeatures.generate_kinematic_features(dataframe)

        # Then, lets break down the entire dataframe into pieces containing data of
        # 1 trajectory in each piece.
        ids_ = list(dataframe.traj_id.value_counts().keys())

        # Get the ideal number of IDs by which the dataframe is to be split.
        df_chunks = []
        for i in range(len(ids_)):
            small_df = ptdf.reset_index().loc[ptdf.reset_index()[const.TRAJECTORY_ID] == ids_[i]]
            df_chunks.append(small_df)

        results = []
        # Here, create 2/3rds number of processes as there are in the system. Some CPUs are
        # kept free at all times in order to not block up the system.
        # (Note: The blocking of system is mostly prevalent in Windows and does not happen very often
        # in Linux. However, out of caution some CPUs are kept free regardless of the system.)
        mp_pool = multiprocessing.Pool(NUM_CPU)
        results = mp_pool.map(helpers.stats_helper, df_chunks)
        mp_pool.close()
        mp_pool.join()

        return pd.concat(results)


"""
    The filters module contains several data filtering functions like
    filtering the data based on time, date, proximity to a point and
    several others.

    | Authors: Yaksh J Haranwala, Salman Haidri
"""
import itertools
import math
import multiprocessing
import os
import warnings
from math import ceil
from typing import Text, Optional

import numpy as np
import pandas as pd

import ptrail.utilities.constants as const
from ptrail.core.TrajectoryDF import PTRAILDataFrame
from ptrail.preprocessing.helpers import Helpers as helper
from ptrail.utilities.exceptions import *

num = os.cpu_count()
NUM_CPU = ceil((num * 2) / 3)


class Filters:
    @staticmethod
    def remove_duplicates(dataframe: PTRAILDataFrame):
        """
        Drop duplicates based on the four following columns:
            1. Trajectory ID
            2. DateTime
            3. Latitude
            4. Longitude

        Duplicates will be dropped only when all the values in the above mentioned
        four columns are the same.

        Returns
        -------
            PTRAILDataFrame
                The dataframe with dropped duplicates.
        """
        return dataframe.reset_index().drop_duplicates(
            subset=[const.DateTime, const.TRAJECTORY_ID, const.LAT, const.LONG],
            keep='first')

    @staticmethod
    def filter_by_traj_id(dataframe: PTRAILDataFrame, traj_id: Text):
        """
            Extract all the trajectory points of a particular trajectory specified
            by the trajectory's ID.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe on which the filtering by ID is to be done.
                traj_id: Text
                    The ID of the trajectory which is to be extracted.

            Returns
            -------
                pandas.core.dataframe.DataFrame
                    The dataframe containing all the trajectory points of the specified trajectory.

            Raises
            ------
                MissingTrajIDException:
                    This exception is raised when the Trajectory ID given by the user does not exist
                    in the dataset.

        """
        to_return = dataframe.reset_index().loc[dataframe.reset_index()[const.TRAJECTORY_ID] == traj_id]
        if len(to_return) > 0:
            return to_return
        else:
            raise MissingTrajIDException(f"{traj_id} is not present in the dataset. "
                                         f"Please check Trajectory ID and try again.")

    @staticmethod
    def get_bounding_box_by_radius(lat: float, lon: float, radius: float):
        """
            Calculates bounding box from a point according to the given radius.

            Parameters
            ----------
                lat: float
                    The latitude of centroid point of the bounding box.
                lon: float
                    The longitude of centroid point of the bounding box.
                radius: float
                    The max radius of the bounding box.
                    The radius is given in metres.

            Returns
            -------
                tuple:
                    The bounding box of the user specified size.

            References
            ----------
                https://mathmesquita.dev/2017/01/16/filtrando-localizacao-em-um-raio.html
        """
        lat, lon = math.radians(lat), math.radians(lon)  # Convert latitude, longitude to radians.

        # Calculate the delta factor for the latitudes and then
        # find the minimum and maximum latitudes.
        latitude_delta = radius / (const.RADIUS_OF_EARTH * 1000)
        lat_one = math.degrees(lat - latitude_delta)
        lat_two = math.degrees(lat + latitude_delta)

        # Calculate the delta factor for the longitudes and then
        # find the minimum and maximum longitudes.
        longitude_delta = math.asin((math.sin(latitude_delta)) / math.cos(lat))
        lon_one = math.degrees(lon - longitude_delta)
        lon_two = math.degrees(lon + longitude_delta)

        # Return the bounding box.
        return (lat_one, lon_one,
                lat_two, lon_two)

    @staticmethod
    def filter_by_bounding_box(dataframe: PTRAILDataFrame, bounding_box: tuple, inside: bool = True):
        """
            Given a bounding box, filter out all the points that are within/outside
            the bounding box and return a dataframe containing the filtered points.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe from which the data is to be filtered out.
                bounding_box: tuple
                    The bounding box which is to be used to filter the data.
                inside: bool
                    Indicate whether the data outside the bounding box is required
                    or the data inside it.

            Returns
            -------
                PTRAILDataFrame
                    The filtered dataframe.
        """
        filt = (
                (dataframe[const.LAT] >= bounding_box[0])
                & (dataframe[const.LONG] >= bounding_box[1])
                & (dataframe[const.LAT] <= bounding_box[2])
                & (dataframe[const.LONG] <= bounding_box[3])
        )
        df = dataframe.loc[filt] if inside else dataframe.loc[~filt]
        return PTRAILDataFrame(df.reset_index(), const.LAT, const.LONG, const.DateTime, const.TRAJECTORY_ID)

    @staticmethod
    def filter_by_date(dataframe: PTRAILDataFrame, start_date: Optional[Text] = None, end_date: Optional[Text] = None):
        """
            Filter the dataset by user-given time range.

            Note
            ----
            | The following options are to be noted for filtering the data:
            |    1. If the start_date and end_date both are not given, then entire
                    dataset itself is returned.
            |    2. If only start_date is given, then the trajectory data after
                    (including the start date) the start date is returned.
            |    3. If only end_date is given, then the trajectory data before
                    (including the end date) the end date is returned.
            |    4. If start_date and end_date both are given then the data between
                    the start_date and end_date (included) are returned.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe that is to be filtered.
                start_date: Optional[Text]
                    The start date from which the points are to be filtered.
                end_date: Optional[Text]
                    The end date before which the points are to be filtered.

            Returns
            -------
                PTRAILDataFrame
                    The filtered dataframe containing the resultant data.

            Raises
            ------
                KeyError:
                    When the Date column is not present in the data.
                ValueError:
                    When the start date is later than the end date.
        """
        try:
            # Reset Index and create a variable for storing filtered DataFrame.
            dataframe = dataframe.reset_index()
            filtered_df = None

            # Convert the user-given string dates to pandas datetime format.
            start_date = pd.to_datetime(start_date) if start_date is not None else None
            end_date = pd.to_datetime(end_date) if end_date is not None else None

            # Case-1: No start and end date are give. Hence just return the original dataframe.
            if start_date is None and end_date is None:
                filtered_df = dataframe

            # Case-2: No start_date is given. Hence, return all the points upto and including
            #         the points on the end date.
            elif start_date is None and end_date is not None:
                filt = dataframe['Date'] <= end_date
                filtered_df = dataframe.loc[filt]

            # Case-3: No end date is given. Hence, return all the point after and including the
            #         points on the start date.
            elif start_date is not None and end_date is None:
                filt = dataframe['Date'] >= start_date
                filtered_df = dataframe.loc[filt]

            # Case-4: Both the start date and end date are given. Hence, return the points between
            #         and including the points on start and end date.
            else:
                if end_date < start_date:
                    raise ValueError(f"End Date should be later than Start Date.")
                else:
                    filt = np.logical_and(dataframe['Date'] >= start_date, dataframe['Date'] <= end_date)
                    filtered_df = dataframe.loc[filt].reset_index()

            # Convert the smaller dataframe back to PTRAILDataFrame and return it.
            return PTRAILDataFrame(filtered_df, const.LAT, const.LONG, const.DateTime, const.TRAJECTORY_ID)

        except KeyError:
            # Ask the user first to create a date.
            raise MissingColumnsException(f"The Date column is missing in the dataset. Please run the "
                                          f"create_date_column() function first before running this filter.")

    @staticmethod
    def filter_by_datetime(dataframe: PTRAILDataFrame, start_dateTime: Optional[Text] = None,
                           end_dateTime: Optional[Text] = None):
        """
            Filter the dataset by user-given time range.

            Note
            ----
            | The following options are to be noted for filtering the data.
            |    1. If the start_dateTime and end_dateTime both are not given, then entire
                    dataset itself is returned.
            |    2. If only start_dateTime is given, then the trajectory data after
                    (including the start datetime) the start date is returned.
            |    3. If only end_dateTime is given, then the trajectory data before
                    (including the end datetime) the end date is returned.
            |    4. If start_dateTime and end_dateTime both are given then the data between
                    the start_dateTime and end_dateTime (included) are returned.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe that is to be filtered.
                start_dateTime: Optional[Text]
                    The start dateTime from which the points are to be filtered.
                end_dateTime: Optional[Text]
                    The end dateTime before which the points are to be filtered.

            Returns
            -------
                PTRAILDataFrame
                    The filtered dataframe containing the resultant data.

            Raises
            ------
                ValueError:
                    When the start datetime is later than the end datetime.
        """
        # Reset Index and create a variable for storing filtered DataFrame.
        dataframe = dataframe.reset_index()
        filtered_df = None

        # Convert the user-given string datetime to pandas datetime format.
        start_dateTime = pd.to_datetime(start_dateTime) if start_dateTime is not None else None
        end_dateTime = pd.to_datetime(end_dateTime) if end_dateTime is not None else None

        # Case-1: No start and end datetime are give. Hence just return the original dataframe.
        if start_dateTime is None and end_dateTime is None:
            filtered_df = dataframe

        # Case-2: No start_dateTime is given. Hence, return all the points upto and including
        #         the points on the end datetime.
        elif start_dateTime is None and end_dateTime is not None:
            filt = dataframe[const.DateTime] <= end_dateTime
            filtered_df = dataframe.loc[filt]

        # Case-3: No end datetime is given. Hence, return all the point after and including the
        #         points on the start datetime.
        elif start_dateTime is not None and end_dateTime is None:
            filt = dataframe[const.DateTime] >= start_dateTime
            filtered_df = dataframe.loc[filt]

        # Case-4: Both the start datetime and end datetime are given. Hence, return the points between
        #         and including the points on start and end datetime.
        else:
            if end_dateTime < start_dateTime:
                raise ValueError(f"End Datetime should be later than Start Datetime.")
            else:
                filt = np.logical_and(dataframe[const.DateTime] >= start_dateTime,
                                      dataframe[const.DateTime] <= end_dateTime)
                filtered_df = dataframe.loc[filt].reset_index()

        # Convert the smaller dataframe back to PTRAILDataFrame and return it.
        return PTRAILDataFrame(filtered_df, const.LAT, const.LONG, const.DateTime, const.TRAJECTORY_ID)

    @staticmethod
    def filter_by_max_speed(dataframe: PTRAILDataFrame, max_speed: float):
        """
            Remove the data points which have speed more than a user given speed.

            Note
            ----
                The max_speed is given in the units m/s (metres per second).

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe which is to be filtered.
                max_speed: float
                    The speed threshold above which the points are to be removed.

            Returns
            -------
                PTRAILDataFrame:
                    PTRAILDataFrame Dataframe containing the resultant dataframe.

            Raises
            ------
                KeyError:
                    The column 'Speed' is not present in the dataframe.
        """
        try:
            dataframe = dataframe.reset_index()
            # Filter out all the values lesser than the max speed
            # The NaNs are filled with max_speed+1 to avoid data loss as well as
            # to avoid false positives in calculation and comparison.
            filt = dataframe['Speed'].fillna(max_speed + 1) <= max_speed
            filtered_df = dataframe.loc[filt].reset_index(drop=True)

            # Convert the smaller dataframe back to PTRAILDataFrame and return it.
            return PTRAILDataFrame(filtered_df, const.LAT, const.LONG, const.DateTime, const.TRAJECTORY_ID)
        except KeyError:
            raise MissingColumnsException(f"The column 'Speed is not present in the dataset. "
                                          f"Please run the function create_speed_from_prev_column() before"
                                          f" running this filter.")

    @staticmethod
    def filter_by_min_speed(dataframe, min_speed: float):
        """
            Remove the data points which have speed less than a user given speed.

            Note
            ----
                The min_speed is given in the units m/s (metres per second).

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe which is to be filtered.
                min_speed: float
                    The speed threshold below which the points are to be removed.

            Returns
            -------
                PTRAILDataFrame:
                    PTRAILDataFrame Dataframe containing the resultant dataframe.

            Raises
            ------
                KeyError:
                    The column 'Speed' is not present in the dataframe.
        """
        try:
            dataframe = dataframe.reset_index()
            # Filter out all the values lesser than the max speed
            # The NaNs are filled with min_speed-1 to avoid data loss as well as
            # to avoid false positives in calculation and comparison.
            filt = dataframe['Speed'].fillna(min_speed - 1) >= min_speed
            filtered_df = dataframe.loc[filt].reset_index()

            # Convert the smaller dataframe back to PTRAILDataFrame and return it.
            return PTRAILDataFrame(filtered_df, const.LAT, const.LONG, const.DateTime, const.TRAJECTORY_ID)
        except KeyError:
            raise MissingColumnsException(f"The column 'Speed is not present in the dataset. "
                                          f"Please run the function create_speed_from_prev_column() before"
                                          f" running this filter.")

    @staticmethod
    def filter_by_min_consecutive_distance(dataframe, min_distance: float):
        """
            Remove the points that have a distance between 2 consecutive points
            lesser than a user specified value.

            Note
            ----
                min_distance is given in metres.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe which is to be filtered.
                min_distance: float
                    The consecutive distance threshold below which the points are to
                    be removed.

            Returns
            -------
                PTRAILDataFrame:
                    The filtered dataframe.

            Raises
            ------
                KeyError:
                    The column 'Distance' is not present in the dataframe.

        """
        try:
            dataframe = dataframe.reset_index()
            # Filter out all the values lesser than the minimum consecutive distance.
            # The NaNs are filled with min_distance-1 to avoid data loss as well as
            # to avoid false positives in calculation and comparison.
            filt = dataframe['Distance'].fillna(min_distance - 1) >= min_distance
            filtered_df = dataframe.loc[filt].reset_index(drop=True)

            # Convert the smaller dataframe back to PTRAILDataFrame and return it.
            return PTRAILDataFrame(filtered_df, const.LAT, const.LONG, const.DateTime, const.TRAJECTORY_ID)
        except KeyError:
            raise MissingColumnsException(f"The column 'Distance is not present in the dataset. "
                                          f"Please run the function create_distance_between_consecutive_column() before"
                                          f" running this filter.")

    @staticmethod
    def filter_by_max_consecutive_distance(dataframe, max_distance: float):
        """
            Remove the points that have a distance between 2 consecutive points
            greater than a user specified value.

            Note
            ----
                max_distance is given in metres.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe which is to be filtered.
                max_distance: float
                    The consecutive distance threshold above which the points are to
                    be removed.

            Returns
            -------
                PTRAILDataFrame:
                    The filtered dataframe.

            Raises
            ------
                KeyError:
                    The column 'Distance' is not present in the dataframe.
            """
        try:
            dataframe = dataframe.reset_index()
            # Filter out all the values greater than the maximum consecutive distance.
            # The NaNs are filled with max_distance+1 to avoid data loss as well as
            # to avoid false positives in calculation and comparison.
            filt = dataframe['Distance'].fillna(max_distance + 1) <= max_distance
            filtered_df = dataframe.loc[filt]

            # Convert the smaller dataframe back to PTRAILDataFrame and return it.
            return PTRAILDataFrame(filtered_df, const.LAT, const.LONG, const.DateTime, const.TRAJECTORY_ID)
        except KeyError:
            raise MissingColumnsException(f"The column 'Distance is not present in the dataset. "
                                          f"Please run the function create_distance_between_consecutive_column() before"
                                          f" running this filter.")

    @staticmethod
    def filter_by_max_distance_and_speed(dataframe, max_distance: float, max_speed: float):
        """
            Filter out values that have distance between consecutive points greater
            than a user-given distance and speed between consecutive points greater than
            a user-given speed

            Note
            ----
                The max_distance is given in metres

            Note
            ----
                The max_speed is given in metres/second (m/s).

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe which is to be filtered.
                max_distance: float
                    The maximum distance between 2 consecutive points.
                max_speed: float
                    The maximum speed between 2 consecutive points.

            Returns
            -------
                PTRAILDataFrame:
                    The filtered dataframe.

            Raises
            ------
                MissingColumnsException:
                    The column 'Distance' or 'Speed' is not present in the dataframe.
        """
        try:
            # filt = Filters.filter_by_max_consecutive_distance(dataframe, max_distance)
            # filtered_df = Filters.filter_by_max_speed(filt, max_speed)

            # Filter the dataframe based on maximum distance and speed.
            filt = np.logical_and(dataframe['Distance'].fillna(max_distance + 1) <= max_distance,
                                  dataframe['Speed'].fillna(max_speed + 1) <= max_speed)
            filtered_df = dataframe.loc[filt]

            return filtered_df  # Return the df filtered on the basis of 2 constraints.
        except KeyError:
            # The system asks the user to only run create_speed_from_prev_column() function because
            # running create_speed_from_prev_column() function create the 'Distance'
            # function automatically if the column is already not present.
            raise MissingColumnsException(f"Either of the columns 'Distance' or 'Speed'"
                                          f"are missing in the dataset. Please run the function "
                                          f"create_speed_column() first before running the function.")

    @staticmethod
    def filter_by_min_distance_and_speed(dataframe, min_distance: float, min_speed: float):
        """
            Filter out values that have distance between consecutive points lesser
            than a user-given distance and speed between consecutive points lesser than
            a user-given speed.

            Note
            ----
                The min_distance is given in metres.

            Note
            ----
                The min_speed is given in metres/second (m/s).

            Parameters
            ----------
            dataframe: PTRAILDataFrame
                The dataframe which is to be filtered.
            min_distance: float
                The minimum distance between 2 consecutive points.
            min_speed: float
                The minimum speed between 2 consecutive points.

            Returns
            -------
                PTRAILDataFrame:
                    The filtered dataframe.

            Raises
            ------
                MissingColumnsException:
                    Either of Distance or Speed column are missing from the dataset.
        """
        try:
            # filt = Filters.filter_by_max_consecutive_distance(dataframe, max_distance)
            # filtered_df = Filters.filter_by_max_speed(filt, max_speed)

            # Filter the dataframe based on minimum distance and speed.
            filt = np.logical_and(dataframe['Distance'] >= min_distance,
                                  dataframe['Speed'] >= min_speed)
            filtered_df = dataframe.loc[filt]

            # Return the df filtered on the basis of 2 constraints.
            return filtered_df
        except KeyError:
            # The system asks the user to only run create_speed_from_prev_column() function because
            # running create_speed_from_prev_column() function create the 'Distance'
            # function automatically if the column is already not present.
            raise MissingColumnsException(f"Either of the columns 'Distance' or 'Speed'"
                                          f"are missing in the dataset. Please run the function "
                                          f"create_speed_column() first before running the function.")

    @staticmethod
    def filter_outliers_by_consecutive_distance(dataframe: PTRAILDataFrame):
        """
            Check the outlier points based on distance between 2 consecutive points.
            Outlier formula:

            |    Lower outlier = Q1 - (1.5*IQR)
            |    Higher outlier = Q3 + (1.5*IQR)
            |    IQR = Inter quartile range = Q3 - Q1

            |    We need to find points between lower and higher outlier

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe which is to be filtered.

            Returns
            -------
                PTRAILDataFrame:
                    The dataframe which has been filtered.

            Raises
            ------
                KeyError:
                    The column 'Distance' is not present in the dataset.
        """
        try:
            # Find the lower and higher quantile first along with the inter-quantile range.
            dataframe = dataframe.reset_index()
            q_low = dataframe['Distance'].quantile(0.25)
            q_high = dataframe['Distance'].quantile(0.75)
            iqr = q_high - q_low
            cut_off = iqr * 1.5  # Cut off value.

            # Now, find the upper limit and the lower limit for the data.
            lower = q_low - cut_off
            higher = q_high + cut_off

            # Filter out the dataframe based on the range calculated and return it.
            df_filt = np.logical_and(dataframe['Distance'] > lower,
                                     dataframe['Distance'] < higher)

            filtered_df = dataframe.loc[df_filt]
            return PTRAILDataFrame(filtered_df, const.LAT, const.LONG, const.DateTime, const.TRAJECTORY_ID)

        except KeyError:
            raise MissingColumnsException(f"The column 'Distance' is missing in the dataset. "
                                          f"Please run the function create_distance_between_consecutive_column() "
                                          f"before running this filter.")

    @staticmethod
    def filter_outliers_by_consecutive_speed(dataframe):
        """
            Check the outlier points based on distance between 2 consecutive points.
            Outlier formula:

            |    Lower outlier = Q1 - (1.5*IQR)
            |    Higher outlier = Q3 + (1.5*IQR)
            |    IQR = Inter quartile range = Q3 - Q1

            |    We need to find points between lower and higher outlier

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe which is to be filtered.

            Returns
            -------
                PTRAILDataFrame:
                    The dataframe which has been filtered.

            Raises
            ------
                KeyError:
                The column 'Distance' is not present in the dataset.
        """
        try:
            q_low = dataframe['Speed'].quantile(0.25)
            q_high = dataframe['Speed'].quantile(0.75)
            iqr = q_high - q_low
            cut_off = iqr * 1.5

            lower = q_low - cut_off
            higher = q_high + cut_off

            df_filt = np.logical_and(dataframe['Speed'] > lower,
                                     dataframe['Speed'] < higher)
            return dataframe.loc[df_filt]

        except KeyError:
            raise MissingColumnsException(f"The column 'Speed' is missing in the dataset. "
                                          f"Please run the function create_speed_from_prev_column() "
                                          f"before running this filter.")

    @staticmethod
    def remove_trajectories_with_less_points(dataframe, num_min_points: Optional[int] = 3):
        """
            Remove out the trajectories from the dataframe which have few points.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe from which trajectories with few points are to be
                    removed.
                num_min_points: Optional[int], default = 2
                    The minimum number of points that a trajectory should have if it
                    is to be retained in the dataset.

            Returns
            -------
                PTRAILDataFrame:
                    The filtered dataframe which does not contain the trajectories
                    with few points anymore.
        """
        dataframe = dataframe.reset_index()
        # Mapping the trajectory id column with only those trajectory ids that contain more
        # or equal points than specified by the user.
        filt = dataframe[const.TRAJECTORY_ID].map(dataframe[const.TRAJECTORY_ID].value_counts()) >= num_min_points

        # Apply the filter, convert the resultant dataframe to PTRAILDataFrame and return it.
        df = dataframe[filt]
        return PTRAILDataFrame(df, const.LAT, const.LONG, const.DateTime, const.TRAJECTORY_ID)

    @staticmethod
    def hampel_outlier_detection(dataframe, column_name: Text):
        """
            Use the hampel filter to remove outliers from the dataset on the basis
            of column specified by the user.

            Warning
            -------
                Do not use Hampel filter outlier detection and try to detect outliers
                with DateTime as it will raise a NotImplementedError as it has not been
                implemented yet by the original author of the Hampel filter.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe from which the outliers are to be removed.
                column_name: Text
                    The column on te basis of which the outliers are to be detected.

            Returns
            -------
                PTRAILDataFrame:
                    The dataframe with the outliers removed.

            Raises
            ------
                KeyError:
                    The user-specified column is not present in the dataset.

            References
            ----------
                Pedrido, M.O., "Hampel", (2020), GitHub repository,
                https://github.com/MichaelisTrofficus/hampel_filter
        """
        # Reset the index of the dataframe and then split the original dataframe into
        # smaller chunks containing a fixed number of trajectory IDs.
        df = dataframe.reset_index()
        df_chunks = helper._df_split_helper(df)

        # create a list using the multiprocessing manager to store the returns of the parallel
        # helper functions.
        processes = [None] * len(df_chunks)
        manager = multiprocessing.Manager()
        return_list = manager.list()

        # Spawn the parallel process and run the hampel filter.
        for i in range(len(processes)):
            processes[i] = multiprocessing.Process(target=Filters._hampel,
                                                   args=(df_chunks[i], column_name, return_list))
            processes[i].start()

        # Join the helper processes again to free up the CPUs and the main memory.
        for j in range(len(processes)):
            processes[j].join()

        warnings.warn("If kinematic features have been generated on the dataframe, then make "
                      "sure to generate them again as outlier detection drops the point from "
                      "the dataframe and does not run the kinematic features again.")

        # Convert the results back to PTRAILDataFrame and return the resultant dataframe.
        return PTRAILDataFrame(pd.concat(return_list),
                               const.LAT, const.LONG, const.DateTime, const.TRAJECTORY_ID)

    @staticmethod
    def _hampel(dataframe, column_name: Text, return_list: list):
        """
            This is a mid-level helper function for the hampel filter. The
            purpose of this function is to split the smaller chunks of dataframe
            received into even smaller chunks containing 1 trajectory ID per chunk,
            then call the helper function from the helper module and run the hampel filter
            on it, then merge all the smaller results and append it to the return list.
        """
        # Split the smaller dataframe further into smaller chunks containing only 1
        # Trajectory ID per index.
        ids_ = list(dataframe[const.TRAJECTORY_ID].value_counts().keys())
        df_chunks = [dataframe.loc[dataframe[const.TRAJECTORY_ID] == ids_[i]] for i in range(len(ids_))]

        # Here, create as many processes at once as there are number of CPUs available in
        # the system - 1. One CPU is kept free at all times in order to not block up
        # the system. (Note: The blocking of system is mostly prevalent in Windows and does
        # not happen very often in Linux. However, out of caution 1 CPU is kept free regardless
        # of the system.)
        small_pool = multiprocessing.Pool(NUM_CPU - 1)
        final = small_pool.starmap(helper.hampel_help,
                                   zip(df_chunks, itertools.repeat(column_name)))
        small_pool.close()
        small_pool.join()

        return_list.append(pd.concat(final))



from .core.TrajectoryDF import PTRAILDataFrame
from .features import helper_functions
from .features import semantic_features
from .features import kinematic_features
from .features import temporal_features
from .preprocessing import filters
from .preprocessing import helpers
from .preprocessing import interpolation
from .utilities import (
    constants,
    conversions,
    DistanceCalculator,
    exceptions,
)

__version__ = "0.4 Beta"



"""
    DistanceCalculator module contains various types of distance formulas that can be used to calculate
    distance between 2 points on the surface of earth depending on the CRS being used.

    | Authors: Salman Haidri, Yaksh J Haranwala
"""

import numpy as np

from ptrail.utilities import constants as const

np.seterr(invalid='ignore')


class FormulaLog:
    @staticmethod
    def haversine_distance(lat1, lon1, lat2, lon2):
        """
            The haversine formula calculates the great-circle distance between 2 points.
            The great-circle distance is the shortest distance over the earth's surface.

            Parameters
            ----------
                lat1: float
                    The latitude value of point 1.
                lon1: float
                    The longitude value of point 1.
                lat2: float
                    The latitude value of point 2.
                lon2: float
                    The longitude value of point 2.

            Returns
            -------
                float
                    The great-circle distance between the 2 points.
        """
        lat1, lon1, lat2, lon2 = np.radians([lat1, lon1, lat2, lon2])
        val_one = (np.sin((lat2 - lat1) / 2.0) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin((lon2 - lon1) / 2.0) ** 2)

        crow_distance = 2 * np.arctan2(val_one ** 0.5, (1 - val_one) ** 0.5)

        return (const.RADIUS_OF_EARTH * crow_distance) * 1000

    @staticmethod
    def bearing_calculation(lat1, lon1, lat2, lon2):
        """
            Calculates bearing between 2 points. Bearing can be defined as direction or
            an angle, between the north-south line of earth or meridian and the line connecting
            the target and the reference point.

            Parameters
            ----------
                lat1:
                    The latitude value of point 1.
                lon1:
                    The longitude value of point 1.
                lat2:
                    The latitude value of point 2.
                lon2:
                    The longitude value of point 2.

            Returns
            -------
                float
                    Bearing between 2 points
        """
        lat1, lon1, lat2, lon2 = np.radians([lat1, lon1, lat2, lon2])

        y = np.cos(lat2) * np.sin(lon2 - lon1)
        x = np.cos(lat1) * np.sin(lat2) - np.sin(lat1) * np.cos(lat2) * np.cos(lon2 - lon1)

        bearing = np.arctan2(y, x)

        return (np.rad2deg(bearing)) % 360.0



"""
    The TrajectoryDF module is the main module containing the PTRAILDataFrame Dataframe
    for storing the Trajectory Data with PTRAIL Library. The Dataframe has
    certain restrictions on what type of data is mandatory in order to be stored as a
    PTRAILDataFrame which is mentioned in the documentation of the constructor.

    | Authors: Yaksh J Haranwala, Salman Haidri
"""
import random
from parser import ParserError
from typing import Dict, List, Union, Optional, Text

import folium
import numpy as np
import pandas as pd
import pandas.core.dtypes.common
from pandas import DataFrame
from pandas._libs import lib

from ptrail.utilities import constants as const
from ptrail.utilities.exceptions import *


class PTRAILDataFrame(DataFrame):
    def __init__(self, data_set: Union[DataFrame, List, Dict], latitude: Text, longitude: Text, datetime: Text,
                 traj_id: Text, rest_of_columns: Optional[List[Text]] = None):
        """
            Construct a trajectory dataframe to store and represent the Trajectory Data.

            Note
            ----
                | The mandatory columns in the dataset are:
                |    1. DateTime
                |    2. Trajectory ID
                |    3. Latitude
                |    4. Longitude

                | ``rest_of_columns`` makes sure that if the data_set is a list, it has appropriate headers
                  that the user wants instead of the default numerical values.

            Parameters
            ----------
                data_set: List, Dictionary or pandas DF.
                    The data provided by the user that needs to be represented and stored.
                datetime: str
                    The header of the datetime column.
                traj_id: str
                    The header of the Trajectory ID column.
                latitude: str
                    The header of the latitude column.
                longitude: str
                    The header of the longitude column.
                rest_of_columns: Optional[list[Text]]
                    A list containing headers of the columns other than the mandatory ones.
        """
        # Case-1: The data is from a dictionary.
        # Here, first check whether the data is in dictionary form and if it is so, then convert into
        # pandas DataFrame first.
        rest_of_columns = [] if rest_of_columns is None else rest_of_columns
        column_list = [latitude, longitude, datetime, traj_id] + rest_of_columns
        if isinstance(data_set, dict):
            data_set = DataFrame.from_dict(data_set)
            data_set = data_set.rename(columns=dict(zip(data_set.columns, [const.LAT, const.LONG,
                                                                           const.DateTime, const.TRAJECTORY_ID])))

        # Case-2: The data is from a list.
        # Here, first check whether the data is in list form and if it is so, then convert into
        # pandas DataFrame first.
        elif isinstance(data_set, list) or isinstance(data_set, np.ndarray):
            data_set = DataFrame(data_set, columns=column_list)

        # Case-3: The data is from a pandas DF.
        # Here, all we have to do is to rename the column names from the data to default names.
        elif isinstance(data_set, DataFrame):
            data_set = self._rename_df_col_headers(data_set, latitude, longitude, datetime, traj_id)

        # Now, renaming the default column names to library default column names.
        column_names = self._get_default_column_names(DateTime=datetime, traj_id=traj_id,
                                                      latitude=latitude, longitude=longitude)
        data_set.rename(columns=column_names, inplace=True)

        # Now checking whether all the columns are present in the data and then verifying the data types
        # of all the columns abd then calling the super() to create and return the dataframe.
        if self._validate_columns(data_set):
            self._validate_data_types(data_set)
            data_set.set_index([const.TRAJECTORY_ID, const.DateTime], inplace=True)
            data_set.sort_values([const.TRAJECTORY_ID, const.DateTime], inplace=True)
            super(PTRAILDataFrame, self).__init__(data_set)

    # ------------------------------ General (Private) Utilities ----------------------------- #
    def _rename_df_col_headers(self, data: DataFrame, lat: Text, lon: Text,
                               datetime: Text, traj_id: Text):
        """
            Change the column headers of the columns when the user given data is in the
            form of a pandas DF while creating the PTRAILDataFrame. This method is mainly
            used when the user reads in data from a csv because the CSV file might
            contain different names for the columns.

            Parameters
            ----------
                data: DataFrame
                    The dataframe whose column names are to be changed.
                lat: Text
                    The header of the Latitude column.
                lon: Text
                    The header of the Longitude column.
                datetime: Text
                    The header of the DateTime column.
                traj_id: Text
                    The header of the Trajectory ID column.

            Returns
            -------
                pandas.DataFrame
                    The pandas dataframe containing the library default column headers.
        """
        cols = data.columns.to_list()  # List of all column names
        # Iterate over the list of column names and check for keywords in the header
        # to help renaming with appropriate terms.
        for i in range(len(cols)):
            if lat in cols[i]:
                cols[i] = const.LAT
            if lon in cols[i]:
                cols[i] = const.LONG
            if datetime in cols[i]:
                cols[i] = const.DateTime
            if traj_id in cols[i]:
                cols[i] = const.TRAJECTORY_ID

        data = data.set_axis(cols, axis=1)  # Change all the column names to modified values.
        return data

    def _validate_data_types(self, data: DataFrame):
        """
            Check whether all the data given by the user is of valid type and if it isn't,
            it converts them to the specified data types.
                1. Trajectory_ID: Any -> str
                2. LAT: Any -> float64
                3. LONG: Any -> float64
                4. DateTime: Any -> dask.datetime64[ns]

            Parameters
            ----------
                data: pd.DataFrame
                    This is the dataframe that contains the data that was passed in by the user.
                    The data is converted to pandas DF eventually in the import_data function anyway.

            Raises
            ------
                KeyError:
                    Dataframe has one of the mandatory columns missing.
                ParserError:
                    The DateTime format provided is invalid and cannot be parsed as pandas DateTime.
                ValueError:
                    One of the data-types cannot be converted.
        """
        try:
            if data.dtypes[const.LAT] != 'float64':
                data[const.LAT] = data[const.LAT].astype('float64')
            if data.dtypes[const.LONG] != 'float64':
                data[const.LONG] = data[const.LONG].astype('float64')
            if data.dtypes[const.DateTime] != 'datetime64[ns]':
                data[const.DateTime] = data[const.DateTime].astype('datetime64[ns]')
            if data.dtypes[const.TRAJECTORY_ID] != 'str':
                data[const.TRAJECTORY_ID] = data[const.TRAJECTORY_ID].astype('str')
        except KeyError:
            raise KeyError('dataframe missing one of lat, lon, datetime columns.')
        except ParserError:
            raise ParserError('DateTime column cannot be parsed')
        except ValueError:
            raise ValueError('dtypes cannot be converted.')

    def _validate_columns(self, data: DataFrame) -> bool:
        """
            Check whether all the mandatory columns are present in the DataFrame or not.

            Parameters
            ----------
                data
                    The DataFrame containing all the data passed in by the user.

            Returns
            -------
                bool
                    Indicate whether or not all the columns that are mandatory are
                    present in the Data given by the user.

            Raises
            ------
                MissingColumnsException
                    One or more of the mandatory columns (Latitude, Longitude, DateTime, Traj_ID)
                    are missing in the data.
        """
        try:
            if np.isin(const.LAT, data.columns) and \
                    np.isin(const.LONG, data.columns) and \
                    np.isin(const.TRAJECTORY_ID, data.columns) and \
                    np.isin(const.DateTime, data.columns):
                return True
        except KeyError:
            raise MissingColumnsException("One of the columns are missing. Please check your data and try again.")

    def _get_default_column_names(self, DateTime, traj_id, latitude, longitude) -> dict:
        """
            Get a dictionary containing the key, value pairs of the library default
            column names for the following columns:
                1. Latitude
                2. Longitude
                3. DateTime
                4. Trajectory ID

            Parameters
            ----------
                DateTime: Text
                    The datetime header of the column already given by the user.
                traj_id: Text
                    The traj_id header of the column already given by the user.
                latitude:
                    The latitude header of the column already given by the user.
                longitude:
                    The longitude header of the column already given by the user.

            Returns
            -------
                dict
                    A dictionary of mappings from the given header by the user to the
                    library default headers.
        """
        return {
            DateTime: const.DateTime,
            traj_id: const.TRAJECTORY_ID,
            latitude: const.LAT,
            longitude: const.LONG
        }

    def set_default_index(self):
        """
            Set the Index of the dataframe back to traj_id and DateTime.

            Raises
            ------
                MissingColumnsException
                    DateTime/traj_id column is missing from the dataset.
        """
        try:
            self.set_index([const.DateTime, const.TRAJECTORY_ID], inplace=True)
        except KeyError:
            raise MissingColumnsException(f"Either of {const.DateTime} or {const.TRAJECTORY_ID} columns are missing.")

    # ------------------------------- Properties ---------------------------------- #
    @property
    def latitude(self):
        """
            Accessor method for the latitude column of the PTRAILDataFrame DataFrame.

            Returns
            -------
                pandas.core.series.Series
                    The Series containing all the latitude values from the DataFrame.

            Raises
            ------
                MissingColumnsException
                    Latitude column is missing from the data.
        """
        try:
            return self[const.LAT]
        except KeyError or IndexError:
            raise MissingColumnsException("The Latitude column is not present in the DataFrame, please verify again.")

    @property
    def longitude(self):
        """
            Accessor method for the longitude column of the PTRAILDataFrame DataFrame.

            Returns
            -------
                pandas.core.series.Series
                    The Series containing all the longitude values from the DataFrame.

            Raises
            ------
                MissingColumnsException
                    Longitude column is missing from the data
        """
        try:
            return self[const.LONG]
        except KeyError or IndexError:
            raise MissingColumnsException("The Longitude column is not present in the DataFrame, please verify again.")

    @property
    def datetime(self):
        """
            Accessor method for the DateTime column of the PTRAILDataFrame DataFrame.

            Returns
            -------
                pandas.core.series.Series
                    The Series containing all the DateTime values from the DataFrame.

            Raises
            ------
                MissingColumnsException
                    DateTime column is missing from the data.
        """
        try:
            return self.index.get_level_values(const.DateTime).to_series()
        except KeyError or IndexError:
            raise MissingColumnsException("The DateTime column is not present in the DataFrame, please verify again.")

    @property
    def traj_id(self):
        """
            Accessor method for the Trajectory_ID column of the DaskTrajectoryDF.

            Returns
            -------
                pandas.core.series.Series
                    The Series containing all the Trajectory_ID values from the DataFrame.

            Raises
            ------
                MissingColumnsException
                    traj_id column is missing from the data.
        """
        try:
            return self.index.get_level_values(const.TRAJECTORY_ID).to_series()
        except KeyError or IndexError:
            raise MissingColumnsException("The Trajectory_ID column is not present in the DataFrame, please verify "
                                          "again.")

    def __str__(self):
        return f"------------------------ Dataset Facts ------------------------------\n\n" \
               f"Number of unique Trajectories in the data: {self.traj_id.nunique()}\n" \
               f"Number of points in the data: {len(self)}\n" \
               f"Dataset time range: {self.datetime.max() - self.datetime.min()}\n" \
               f"Datatype of the DataFrame: {type(self)}\n" \
               f"Dataset Bounding Box:" \
               f" {(self.latitude.min(), self.longitude.min(), self.latitude.max(), self.longitude.max())}\n\n" \
               f"---------------------------------------------------------------------"

    # ------------------------------- File and DF Operations ----------------------------- #

    def to_numpy(self, dtype=None, copy: bool = False, na_value=lib.no_default) -> np.ndarray:
        """
            Convert the DataFrame to a NumPy array.By default, the dtype of the returned array will
            be the common dtype of all types in the DataFrame. For example, if the dtypes are float16
            and float32, the results dtype will be float32. This may require copying data and coercing
            values, which may be expensive

            Parameters
            ----------
                dtype:
                    The dtype to pass to :meth:`numpy.asarray`.
                copy:
                    Whether to ensure that the returned value is not a view on another array.
                    Note that ``copy=False`` does not *ensure* that ``to_numpy()`` is no-copy.
                    Rather, ``copy=True`` ensure that a copy is made, even if not strictly necessary.
                na_value:
                    The value to use for missing values. The default value depends on `dtype` and the
                    dtypes of the DataFrame columns.
        """
        return self.reset_index(drop=False).to_numpy()

    def sort_by_traj_id_and_datetime(self, ascending=True):
        """
            Sort the trajectory in Ascending or descending order based on the following 2
            columns in order:
                1. Trajectory ID
                2. DateTime

            Parameters
            ----------
                ascending: bool
                    Whether to sort the values in ascending order or descending order.

            Returns
            -------
                PTRAILDataFrame
                    The sorted dataframe.
        """
        return self.sort_values([const.TRAJECTORY_ID, const.DateTime], ascending=ascending)

    # ------------------------------------- Visualization ----------------------------------- #
    def plot_folium_traj(self, weight: float = 3, opacity: float = 0.8):
        """
            Use folium to plot the trajectory on a map.

            Parameters
            ----------
                weight: float
                    The weight of the trajectory line on the map.
                opacity: float
                    The opacity of the trajectory line on the map.

            Returns
            -------
                folium.folium.Map
                    The map with plotted trajectory.
        """
        sw = self[['lat', 'lon']].min().values.tolist()
        ne = self[['lat', 'lon']].max().values.tolist()

        # Create a map with the initial point.
        map_ = folium.Map(location=(self.latitude[0], self.longitude[0]))

        ids_ = list(self.traj_id.value_counts().keys())
        colors = ["#" + ''.join([random.choice('123456789BCDEF') for j in range(6)])
                  for i in range(len(ids_))]

        for i in range(len(ids_)):
            # First, filter out the smaller dataframe.
            small_df = self.reset_index().loc[self.reset_index()[const.TRAJECTORY_ID] == ids_[i],
                                              [const.LAT, const.LONG]]

            # Then, create (lat, lon) pairs for the data points.
            locations = []
            for j in range(len(small_df)):
                locations.append((small_df['lat'].iloc[j], small_df['lon'].iloc[j]))

            # Create popup text for the

            # Create start and end markers for the trajectory.
            folium.Marker([small_df['lat'].iloc[0], small_df['lon'].iloc[0]],
                          color='green',
                          popup=f'Trajectory ID: {ids_[i]} \n'
                                f'Latitude: {locations[0][0]} \n'
                                f'Longitude: {locations[0][1]}',
                          marker_color='green',
                          icon=folium.Icon(icon_color='green', icon=None)).add_to(map_)

            folium.Marker([small_df['lat'].iloc[-1], small_df['lon'].iloc[-1]],
                          color='green',
                          popup=f'Trajectory ID: {ids_[i]} \n'
                                f'Latitude: {locations[-1][0]} \n'
                                f'Longitude: {locations[-1][1]}',
                          marker_color='red',
                          icon=folium.Icon(icon_color='red', icon=None)).add_to(map_)

            # Add trajectory to map.
            folium.PolyLine(locations,
                            color=colors[i],
                            weight=weight,
                            opacity=opacity).add_to(map_)

        map_.fit_bounds([sw, ne])
        return map_



from setuptools import find_packages
from setuptools import setup

with open('README.md', 'r') as f:
    LONG_DESCRIPTION = f.read()

REQUIRED_PKGS = ['numpy>=1.20',
                 'hampel >= 0.0.5',
                 'pandas >= 1.2.5',
                 'scipy >= 1.6.2',
                 'folium >= 0.12',
                 'matplotlib >= 3.3.4',
                 'scikit-learn >= 0.24.2',
                 'osmnx >= 1.1.1',
                 'geopandas >= 0.8.1',
                 'shapely >= 1.7.1',
                 ]

setup(
    name='ptrail',
    packages=find_packages(),
    version='0.4 Beta',
    license='new BSD',
    python_requires='>=3.6',
    description='PTRAIL: A Mobility-data Preprocessing Library using parallel computation.',
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    maintainer='PTRAIL Developers',
    maintainer_email='mobilitylab2021@gmail.com',
    classifiers=['Intended Audience :: Science/Research',
                 'Intended Audience :: Developers',
                 'License :: OSI Approved',
                 'Programming Language :: Python',
                 'Topic :: Software Development',
                 'Topic :: Scientific/Engineering',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python :: 3',
                 ],
    install_requires=REQUIRED_PKGS,
    url='https://github.com/YakshHaranwala/PTRAIL.git',
    include_package_data=True,
)



"""
    This file contains all the custom designed exception headers.
    There is nothing here but the exception headers and pass written inside them.
    The purpose of the file is to store all exceptions in one place.

    | Authors: Yaksh J Haranwala, Salman Haidri
"""


class NoHeadersException(Exception):
    pass


class MissingColumnsException(Exception):
    pass


class DataTypeMismatchException(Exception):
    pass


class MandatoryColumnException(Exception):
    pass


class MissingTrajIDException(Exception):
    pass


class NotAllowedError(Exception):
    pass



"""
    This module contains all the helper functions for the parallel calculations in
    the spatial and temporal features classes.

    Warning
    -------
        These functions should not be used directly as they would result in a
        slower calculation and execution times. In some cases, these functions
        might even yield wrong results if used directly. They are meant to be used
        only as helpers. For calculation of features, use the ones in the
        features package.

    | Authors: Yaksh J Haranwala, Salman Haidri
"""
import os
from math import ceil

import numpy as np
import pandas as pd

from ptrail.utilities import constants as const
from ptrail.utilities.DistanceCalculator import FormulaLog as calc


class Helpers:
    # ------------------------------------ Temporal Helpers --------------------------------------#
    @staticmethod
    def traj_duration_helper(dataframe, ids_):
        """
            Calculate the duration of the trajectory i.e. subtract the max time of
            the trajectory by the min time of the trajectory.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe containing all the original data.
                ids_: list
                    A list containing all the Trajectory IDs present in the dataset.

            Returns
            -------
                pandas.core.dataframe.DataFrame
                    The resultant dataframe containing all the trajectory durations.
        """
        durations = []  # A list for storing results.

        # Iterate over each ID and calculate the time duration of each unique ID.
        for i in range(len(ids_)):
            # Filter out only the points of the ID in question.
            small = dataframe.loc[dataframe[const.TRAJECTORY_ID] == ids_[i], [const.DateTime]]

            # Calculate the duration of the trajectory in question nd append
            # a row containing [traj_duration, traj_id] to the results list.
            durations.append([(small.max() - small.min())[0], ids_[i]])

        # Convert the list containing results to a pandas dataframe, reset the index
        # and then rename the columns.
        result = pd.DataFrame(durations).reset_index(drop=True).rename(columns={0: "Traj_Duration",
                                                                                1: const.TRAJECTORY_ID})
        # Set the index to traj_id and return it.
        return result.set_index(const.TRAJECTORY_ID)

    @staticmethod
    def start_time_helper(dataframe, ids_):
        """
            This function is the helper function of the get_start_time(). The get_start_time() function
            delegates the task of calculating the start_time of the trajectories in the dataframe because the
            original functions runs multiple instances of this function in parallel. This function finds the start
            time of the specified trajectory IDs the DF and then another returns dataframe containing
            start latitude, start longitude, DateTime and trajectory ID for each trajectory

            Parameter
            ---------
                dataframe: PTRAILDataFrame
                    The dataframe containing the original data.
                ids_: list
                    List of trajectory ids for which the start times are to be calculated

            Returns
            -------
                pandas.core.dataframe.Dataframe
                    New dataframe containing Trajectory ID as index and start time of all
                    trajectories.
        """
        results = []

        # Loops over the length of trajectory ids. Filter the dataframe according to each of the ids
        # and then further filter that dataframe according to the earliest(minimum) time.
        # And then append the data of that earliest time into a list.
        for i in range(len(ids_)):
            filt = (dataframe.loc[dataframe[const.TRAJECTORY_ID] == ids_[i],
                                  [const.DateTime, const.LAT, const.LONG]])
            start_time = (filt.loc[filt[const.DateTime] == filt[const.DateTime].min()]).reset_index()
            results.append([start_time[const.DateTime][0], ids_[i]])

        # Make a new dataframe containing Latitude Longitude and Trajectory id
        df = pd.DataFrame(results).reset_index(drop=True).rename(columns={0: const.DateTime,
                                                                          1: const.TRAJECTORY_ID})
        # Return the dataframe by setting Trajectory id as index
        return df.set_index(const.TRAJECTORY_ID)

    @staticmethod
    def end_time_helper(dataframe, ids_):
        """
            This function is the helper function of the get_end_time(). The get_end_time() function
            delegates the task of calculating the end_time of the trajectories in the dataframe because the
            original functions runs multiple instances of this function in parallel. This function finds the end
            time of the specified trajectory IDs the DF and then another returns dataframe containing
            end latitude, end longitude, DateTime and trajectory ID for each trajectory

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe containing the original data.
                ids_: list
                    List of trajectory ids for which the end times are to be calculated

            Returns
            -------
                pandas.core.dataframe.Dataframe
                    New dataframe containing Trajectory ID as index end time of all trajectories.
        """
        results = []

        # Loops over the length of trajectory ids. Filter the dataframe according to each of the ids
        # and then further filter that dataframe according to the latest(maximum) time.
        # And then append the data of that latest time into a list.
        for i in range(len(ids_)):
            filt = (dataframe.loc[dataframe[const.TRAJECTORY_ID] == ids_[i],
                                  [const.DateTime, const.LAT, const.LONG]])
            start_time = (filt.loc[filt[const.DateTime] == filt[const.DateTime].max()]).reset_index()
            results.append([start_time[const.DateTime][0], ids_[i]])

        # Make a new dataframe containing Latitude Longitude and Trajectory id
        df = pd.DataFrame(results).reset_index(drop=True).rename(columns={0: const.DateTime,
                                                                          1: const.TRAJECTORY_ID})
        # Return the dataframe by setting Trajectory id as index
        return df.set_index(const.TRAJECTORY_ID)

    # -------------------------------------- Spatial Helpers ----------------------------------------------- #
    @staticmethod
    def distance_between_consecutive_helper(dataframe):
        """
            This function is the helper function of the create_distance_between_consecutive_column() function.
            The create_distance_between_consecutive_column() function delegates the actual task of calculating
            the distance between 2 consecutive points. This function does the calculation and creates a column
            called Distance_prev_to_curr and places it in the dataframe and returns it.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe on which calculation is to be performed.

            Returns
            -------
                core.TrajectoryDF.PTRAILDataFrame
                    The dataframe containing the resultant Distance_prev_to_curr column.

            References
            ----------
                Arina De Jesus Amador Monteiro Sanches. 'Uma Arquitetura E Imple-menta ̧c ̃ao
                Do M ́odulo De Pr ́e-processamento Para Biblioteca Pymove'.Bachelor’s thesis.
                Universidade Federal Do Cear ́a, 2019.
        """
        # Reset the index and set it to trajectory ID in order to iterate
        # over the dataframe based on trajectory ID.
        dataframe = dataframe.reset_index().set_index(const.TRAJECTORY_ID)
        ids_ = dataframe.index.unique()  # Find out all the unique IDs in the dataframe.

        # For each unique ID, calculate the haversine distance and then assign it to
        # a new column and add that column to the dataframe.
        for val in ids_:
            curr_lat = dataframe.at[val, 'lat']
            curr_lon = dataframe.at[val, 'lon']
            size_id = curr_lat.size

            # Check whether the IDs are changing in the dataset and if they are, then assign
            # Nan as the value to the first point of the new Trajectory ID.
            if size_id <= 1:
                dataframe.at[val, 'Distance'] = np.nan
            else:
                prev_lat = curr_lat.shift(1)
                prev_lon = curr_lon.shift(1)
                dataframe.at[val, 'Distance'] = \
                    calc.haversine_distance(prev_lat, prev_lon, curr_lat, curr_lon)

        return dataframe.reset_index()  # Reset the index and return the dataframe.

    @staticmethod
    def distance_from_start_helper(dataframe):
        """
            This function is the helper function of the create_distance_from_start_column() function.
            The create_distance_from_start_column() function delegates the actual task of calculating
            the distance between 2 the start point of the trajectory to the current point.This function
            does the calculation and creates a column called Distance_start_to_curr and places it in the
            dataframe and returns it.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe on which calculation is to be performed.

            Returns
            -------
                pandas.core.dataframe
                    The dataframe containing the resultant Distance_start_to_curr column.
        """
        dataframe = dataframe.reset_index().set_index(const.TRAJECTORY_ID)
        ids_ = dataframe.index.unique()  # Find out all the unique IDs in the dataframe.

        # For each unique ID, calculate the haversine distance and then assign it to
        # a new column and add that column to the dataframe.
        for val in ids_:
            curr_lat = dataframe.at[val, const.LAT]
            curr_lon = dataframe.at[val, const.LONG]
            size_id = curr_lat.size

            # Here, the purpose of the if else statement is as follows:
            #   In the above curr_lat variable, when the current latitude is extracted, sometimes
            #   if the trajectory has only a single point, then it only returns a single float value
            #   which naturally cannot be indexed. This if else statement checks if the value returned
            #   by the curr_lat extraction is a float and if so, then dont try to index it and instead
            #   take the value itself. The values are also shifted by 1 to avoid the distance calculation
            #   yielding the value 0 and instead give out NaN.
            start_lat = pd.Series(np.full(size_id,
                                          curr_lat[0] if type(curr_lat) is not np.float64 else curr_lat)).shift(1)
            start_lon = pd.Series(np.full(size_id,
                                          curr_lon[0] if type(curr_lat) is not np.float64 else curr_lat)).shift(1)

            # Check whether the IDs are changing in the dataset and if they are, then assign
            # Nan as the value to the first point of the new Trajectory ID.
            if size_id <= 1:
                dataframe.at[val, 'Distance_from_start'] = np.nan

            else:
                dataframe.at[val, 'Distance_from_start'] = \
                    calc.haversine_distance(start_lat, start_lon, curr_lat, curr_lon)

        return dataframe.reset_index()  # Reset the index and return the dataframe.

    @staticmethod
    def distance_from_given_point_helper(dataframe, coordinates):
        """
            This function is the helper function of the create_distance_from_point() function. The
            create_distance_from_point() function delegates the actual task of calculating distance
            between the given point to all the points in the dataframe to this function. This function
            calculates the distance and creates another column called 'Distance_to_specified_point'.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe on which calculation is to be done.
                coordinates: tuple
                    The coordinates from which the distance is to be calculated.

            Returns
            -------
                pandas.core.dataframe.DataFrame
                    The dataframe containing the resultant Distance_from_(x, y) column.
        """
        # First, lets fetch the latitude and longitude columns from the dataset and store it
        # in a numpy array.
        latitudes = np.array(dataframe[const.LAT])
        longitudes = np.array(dataframe[const.LONG])
        distances = np.zeros(len(latitudes))

        # Now, lets calculate the Great-Circle (Haversine) distance between the 2 points and store
        # each of the values in the distance numpy array.
        for i in range(len(latitudes)):
            distances[i] = calc.haversine_distance(coordinates[0], coordinates[1],
                                                   latitudes[i], longitudes[i])

        dataframe[f'Distance_from_{coordinates}'] = distances
        return dataframe

    @staticmethod
    def point_within_range_helper(dataframe, coordinates, dist_range):
        """
            This is the helper function for create_point_within_range() function. The
            create_point_within_range_column()

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe on which the operation is to be performed.
                coordinates: tuple
                    The coordinates from which the distance is to be checked.
                dist_range:
                    The range within which the distance from the coordinates should lie.

            Returns
            -------
                pandas.core.dataframe.DataFrame
                    The dataframe containing the resultant Within_X_m_from_(x,y) column.
        """
        # First, lets fetch the latitude and longitude columns from the dataset and store it
        # in a numpy array.
        latitudes = np.array(dataframe[const.LAT])
        longitudes = np.array(dataframe[const.LONG])
        distances = []

        # Now, lets calculate the Great-Circle (Haversine) distance between the 2 points and then check
        # whether the distance is within the user specified range and store each of the values in the \
        # distance numpy array.
        for i in range(len(latitudes)):
            distances.append(calc.haversine_distance(coordinates[0], coordinates[1],
                                                     latitudes[i], longitudes[i]) <= dist_range)

        # Now, assign the column containing the results calculated above and
        # return the dataframe.
        dataframe[f'Within_{dist_range}_m'] = distances
        return dataframe

    @staticmethod
    def bearing_helper(dataframe):
        """
            This function is the helper function of the create_bearing_column(). The create_bearing_column()
            delegates the task of calculation of bearing between 2 points to this function because the original
            functions runs multiple instances of this function in parallel. This function does the calculation
            of bearing between 2 consecutive points in the entire DF and then creates a column in the dataframe
            and returns it.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe on which the calculation is to be done.

            Returns
            -------
                PTRAILDataFrame:
                    The dataframe containing the Bearing column.
        """
        # Reset the index and set it to trajectory ID in order to iterate
        # over the dataframe based on trajectory ID.
        dataframe = dataframe.reset_index().set_index(const.TRAJECTORY_ID)
        ids_ = dataframe.index.unique()  # Find out all the unique IDs in the dataframe.

        # For each unique ID, calculate the haversine distance and then assign it to
        # a new column and add that column to the dataframe.
        for val in ids_:
            curr_lat = dataframe.at[val, 'lat']
            curr_lon = dataframe.at[val, 'lon']
            size_id = curr_lat.size

            # Check whether the IDs are changing in the dataset and if they are, then assign
            # Nan as the value to the first point of the new Trajectory ID.
            if size_id <= 1:
                dataframe.at[val, 'Bearing'] = np.nan
            else:
                prev_lat = curr_lat.shift(1)
                prev_lon = curr_lon.shift(1)
                dataframe.at[val, 'Bearing'] = \
                    calc.bearing_calculation(prev_lat, prev_lon, curr_lat, curr_lon)

        return dataframe.reset_index()

    @staticmethod
    def start_location_helper(dataframe, ids_):
        """
            This function is the helper function of the get_start_location(). The get_start_location() function
            delegates the task of calculating the start location of the trajectories in the dataframe because the
            original functions runs multiple instances of this function in parallel. This function finds the start
            location of the specified trajectory IDs the DF and then another returns dataframe containing start
            latitude, start longitude and trajectory ID for each trajectory


            Parameter
            ---------
                dataframe: PTRAILDataFrame
                    The dataframe of which the locations are to be found.dataframe
                ids_: list
                    List of trajectory ids for which the start locations are to be calculated

            Returns
            -------
                pandas.core.dataframe.Dataframe
                    New dataframe containing Trajectory as index and latitude and longitude
        """
        results = []

        # Loops over the length of trajectory ids. Filter the dataframe according to each of the ids
        # and then further filter that dataframe according to the earliest(minimum) time.
        # And then append the start location of that earliest time into a list
        for i in range(len(ids_)):
            filt = (dataframe.loc[dataframe[const.TRAJECTORY_ID] == ids_[i],
                                  [const.DateTime, const.LAT, const.LONG]])
            start_loc = (filt.loc[filt[const.DateTime] == filt[const.DateTime].min(),
                                  [const.LAT, const.LONG]]).reset_index()
            results.append([start_loc[const.LAT][0], start_loc[const.LONG][0], ids_[i]])

        # Make a new dataframe containing Latitude Longitude and Trajectory id
        df = pd.DataFrame(results).reset_index(drop=True).rename(columns={0: const.LAT,
                                                                          1: const.LONG,
                                                                          2: const.TRAJECTORY_ID})

        # Return the dataframe by setting Trajectory id as index
        return df.set_index(const.TRAJECTORY_ID)

    @staticmethod
    def end_location_helper(dataframe, ids_):
        """
            This function is the helper function of the get_end_location(). The get_end_location() function
            delegates the task of calculating the end location of the trajectories in the dataframe because the
            original functions runs multiple instances of this function in parallel. This function finds the end
            location of the specified trajectory IDs the DF and then another returns dataframe containing
            end latitude, end longitude and trajectory ID for each trajectory

            Parameter
            ---------
                dataframe: PTRAILDataFrame
                    The dataframe of which the locations are to be found.dataframe
                ids_: list
                    List of trajectory ids for which the end locations are to be calculated

            Returns
            -------
                pandas.core.dataframe.Dataframe
                    New dataframe containing Trajectory ID as index and latitude and longitude
                    as other 2 columns.
        """
        results = []

        # Loops over the length of trajectory ids. Filter the dataframe according to each of the ids
        # and then further filter that dataframe according to the latest(maximum) time.
        # And then append the end location of that latest time into a list.
        for i in range(len(ids_)):
            filt = (dataframe.loc[dataframe[const.TRAJECTORY_ID] == ids_[i],
                                  [const.DateTime, const.LAT, const.LONG]])
            start_loc = (filt.loc[filt[const.DateTime] == filt[const.DateTime].max(),
                                  [const.LAT, const.LONG]]).reset_index()
            results.append([start_loc[const.LAT][0], start_loc[const.LONG][0], ids_[i]])

        # Make a new dataframe containing Latitude Longitude and Trajectory id
        df = pd.DataFrame(results).reset_index(drop=True).rename(columns={0: const.LAT,
                                                                          1: const.LONG,
                                                                          2: const.TRAJECTORY_ID})
        # Return the dataframe by setting Trajectory id as index
        return df.set_index(const.TRAJECTORY_ID)

    @staticmethod
    def number_of_location_helper(dataframe, ids_):
        """
            This is the helper function for the get_number_of_locations() function. The
            get_number_of_locations() delegates the actual task of calculating the number of
            unique locations visited by a particular object to this function. This function
            calculates the number of unique locations by each of the unique object and returns
            a dataframe containing the results.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                    The dataframe containing all the original data.
                ids_: list
                    The list of ids for which the number of unique locations visited
                    is to be calculated.

            Returns
            -------
                pandas.core.dataframe.DataFrame
                    dataframe containing the results.
        """
        results = []  # A list for storing results.

        # Iterate over each ID and calculate the number of locations visited by each ID.
        for i in range(len(ids_)):
            # Filter out only the points of the ID in question.
            filt = (dataframe.loc[dataframe[const.TRAJECTORY_ID] == ids_[i],
                                  [const.DateTime, const.LAT, const.LONG]])

            # Calculate the total number of unique (lat, lon) points visited
            # by the ID and append a row containing [# of unique locations, traj_id]
            # to the results list.
            results.append([filt.groupby([const.LAT, const.LONG]).ngroups, ids_[i]])

        # Convert the list containing results to a pandas dataframe, reset the index
        # and then rename the columns.
        df = pd.DataFrame(results).reset_index(drop=True).rename(columns={0: "Number of Unique Coordinates",
                                                                          1: const.TRAJECTORY_ID})
        # Set the index to traj_id and return it.
        return df.set_index(const.TRAJECTORY_ID)

    @staticmethod
    def stats_helper(df):
        new_df = df.reset_index()[['traj_id', 'Distance', 'Distance_from_start', 'Speed',
                                   'Acceleration', 'Jerk', 'Bearing', 'Bearing_Rate',
                                   'Rate_of_bearing_rate']]
        stats = new_df.reset_index(drop=True).describe(percentiles=[0.1, 0.25, 0.5, 0.75, 0.9]).transpose()
        stats['traj_id'] = new_df['traj_id'].iloc[0]
        stats = stats[['traj_id', 'mean', 'std', 'min', '10%',
                       '25%', '50%', '75%', '90%', 'max']]
        return stats.reset_index().rename(columns={'index': 'Columns'}).set_index(['traj_id', 'Columns'])

    # ------------------------------------ Semantic Helpers ------------------------------------- #
    @staticmethod
    def visited_poi_helper(df, surrounding_data, dist_column_label, nearby_threshold):
        """
            Given a Trajectory dataframe and another dataset with the surrounding data,
            find whether the given object is nearby a point of interest or not.

            Parameters
            ----------
                df:
                    The dataframe containing the trajectory data.
                surrounding_data:
                    The dataframe containing the data of the surroundings.
                dist_column_label: Text
                    The label of the column containing the distance of the coords from
                    the nearest POI.
                nearby_threshold: int
                    The maximum distance between the POI and the current location of the object
                    within which the object is considered to be crossing/visiting the POI.

            Returns
            -------
                The original dataframe with another column added to it indicating whether
                each point is within
        """
        # A boolean list to store if that point in trajectory lies around a POI.
        POI = []
        df2 = surrounding_data.copy()

        try:
            # Loop for every point in the dataframe and create a distance column with distance from every point in the
            # surrounding data. Then use this distance column and the distance of POI column and compare each value.
            # Store the comparison in an array True if they lie within the threshold else false. And if any of the
            # value in it is true then that point in the dataframe was near a POI.
            for i in range(len(df)):
                dist_array = Helpers.distance_from_given_point_helper(df2, (df['lat'][i], df['lon'][i]))[
                    f'Distance_from_{df["lat"][i], df["lon"][i]}'].to_numpy()
                poi_array = df2[dist_column_label].to_numpy()

                dist_comp = np.abs(poi_array - dist_array) <= nearby_threshold
                POI.append(np.any(dist_comp))

            # Append the boolean list containing whether each point was near the POI of interest or not
            df['Nearby_POI'] = POI

            return df
        except KeyError:
            raise KeyError(f"The column {dist_column_label} does not exist in the dataset.")

    # ------------------------------------ General Utilities ------------------------------------ #
    @staticmethod
    def _get_partition_size(size):
        """
            Takes number of ids and makes use of a formula that gives a factor to makes set of ids
            according to the number of processors available to work with.

            Parameters
            ----------
                size: int
                    The total number of trajectory IDs in the dataset.

            Returns
            -------
                int
                   The factor by which the datasets are to be split.
        """
        # Based on the Operating system, get the number of CPUs available for
        # multiprocessing.
        num = os.cpu_count()
        num = int(num)
        NUM_CPU = ceil((num * 2) / 3)

        # Integer divide the total number of Trajectory IDs by the number of available CPUs
        # and square the number because if too many partitions are made, then it does more
        # harm than good for the execution speed. The factor of 1 is added to avoid errors
        # when the integer division yields a 0.
        factor = ((size // NUM_CPU) ** 2) + 1

        # Return the factor if it is less than 100 otherwise return 100.
        # This factor hence is capped at 100.
        return factor if factor < 100 else 100

    @staticmethod
    def _df_split_helper(dataframe):
        """
            This is the helper function for splitting up dataframes into smaller chunks.
            This function is widely used for main functions to help split the original
            dataframe into smaller chunks based on a fixed range of IDs. This function
            splits the dataframes based on a predetermined number, stores them in a list
            and returns it.

            Note
            ----
                The dataframe is split based on the number of CPU cores available for.
                For more info, take a look at the documentation of the get_partition_size()
                function.

            Parameters
            ----------
                dataframe: PTRAILDataFrame
                The dataframe that is to be split.

            Returns
            -------
                list:
                    The list containing smaller dataframe chunks.
        """
        # First, create a list containing all the ids of the data and then further divide that
        # list items and split it into sub-lists of ids equal to split_factor.
        ids_ = list(dataframe.traj_id.value_counts().keys())

        # Get the ideal number of IDs by which the dataframe is to be split.
        split_factor = Helpers._get_partition_size(len(ids_))
        ids_ = [ids_[i: i + split_factor] for i in range(0, len(ids_), split_factor)]

        # Now split the dataframes based on set of Trajectory ids.
        # As of now, each smaller chunk is supposed to have data of 100
        # trajectory IDs max
        df_chunks = [dataframe.loc[dataframe.index.get_level_values(const.TRAJECTORY_ID).isin(ids_[i])]
                     for i in range(len(ids_))]
        return df_chunks

    # @staticmethod
    # def _partition_size(size, cpu_count):
    #     """
    #         Takes number of ids and makes use of a formula that gives a factor to makes set of ids
    #         according to the number of processors available to work with.
    #
    #         Parameters
    #         ----------
    #             size: int
    #                 The total number of trajectory IDs in the dataset.
    #
    #         Returns
    #         -------
    #             int
    #                The factor by which the datasets are to be split.
    #     """
    #     # Based on the Operating system, get the number of CPUs available for
    #     # multiprocessing.
    #     available_cpus = len(os.sched_getaffinity(0)) if os.name == 'posix' \
    #         else psutil.cpu_count()  # Number of available CPUs.
    #
    #     # Integer divide the total number of Trajectory IDs by the number of available CPUs
    #     # and square the number because if too many partitions are made, then it does more
    #     # harm than good for the execution speed. The factor of 1 is added to avoid errors
    #     # when the integer division yields a 0.
    #     factor = ((size // cpu_count)) + 1
    #
    #     # Return the factor if it is less than 100 otherwise return 100.
    #     # This factor hence is capped at 100.
    #     return factor if factor < 100 else 100
